#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define TILEWIDTH ( 16 )

#define RECTA ( 1 )
#define RECTB ( 4 )

#define BLK_A ( TILEWIDTH*RECTA )
#define BLK_B ( TILEWIDTH*RECTB )

static texture<int2, 1> texture_Ah, texture_Al;
static texture<int2, 1> texture_Bh, texture_Bl;

static __inline__ __device__
double fetch_x_Ah( const int& i )
{
  register int2 v = tex1Dfetch( texture_Ah, i );

  return __hiloint2double( v.y, v.x );
}

static __inline__ __device__
double fetch_x_Al( const int& i )
{
  register int2 v = tex1Dfetch( texture_Al, i );

  return __hiloint2double( v.y, v.x );
}

static __inline__ __device__
double fetch_x_Bh( const int& i )
{
  register int2 v = tex1Dfetch( texture_Bh, i );

  return __hiloint2double( v.y, v.x );
}

static __inline__ __device__
double fetch_x_Bl( const int &i )
{
  register int2 v = tex1Dfetch( texture_Bl, i );

  return __hiloint2double( v.y, v.x );
}


// construct a full matrix from lower part of symmetric matrix
static __global__
void gddSYLtoFull
( int m, double Atrih[], double Atril[],
  double Afullh[], double Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix>=iy ){
      int i = ix+iy*lda;
      Afullh[i] = Atrih[i];
      Afulll[i] = Atril[i];
    }
    else{
      int i = ix+iy*lda;
      int j = iy+ix*lda;
      Afullh[i] = Atrih[j];
      Afulll[i] = Atril[j];
    }
  }
}

// construct a full matrix from upper part of symmetric matrix
static __global__
void gddSYUtoFull
( int m, double Atrih[], double Atril[],
  double Afullh[], double Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix<=iy ){
      int i = ix+iy*lda;
      Afullh[i] = Atrih[i];
      Afulll[i] = Atril[i];
    }
    else{
      int i = ix+iy*lda;
      int j = iy+ix*lda;
      Afullh[i] = Atrih[j];
      Afulll[i] = Atril[j];
    }
  }
}

// construct a full matrix from lower part of triangular matrix
static __global__
void gddTLtoFull
( int m, double Atrih[], double Atril[],
  double Afullh[], double Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    int i = ix+iy*lda;
    Afullh[i] = Atrih[i] * double(ix>=iy);
    Afulll[i] = Atril[i] * double(ix>=iy);
  }
}

// construct a full matrix from upper part of triangular matrix
static __global__
void gddTUtoFull
( int m, double Atrih[], double Atril[],
  double Afullh[], double Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    int i = ix+iy*lda;
    Afullh[i] = Atrih[i] * double(ix<=iy);
    Afulll[i] = Atril[i] * double(ix<=iy);
  }
}


// construct a full matrix from lower part of unit-diagonal triangular matrix
static __global__
void gddTLDtoFull
( int m, double Atrih[], double Atril[],
  double Afullh[], double Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    int i = ix+iy*lda;
    Afullh[i] = Atrih[i] * double(ix>iy) + double(ix==iy);
    Afulll[i] = Atril[i] * double(ix>iy);
  }
}

// construct a full matrix from upper part of unit-diagonal triangular matrix
static __global__
void gddTUDtoFull
( int m, double Atrih[], double Atril[],
  double Afullh[], double Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    int i = ix+iy*lda;
    Afullh[i] = Atrih[i] * double(ix<iy) + double(ix==iy);
    Afulll[i] = Atril[i] * double(ix<iy);
  }
}

// construct a full matrix from lower part of triangular matrix
static __global__ void
gddFulltoFull
( int m,int n, double Ah[], double Al[],
  double Bh[], double Bl[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<n ){
    int i = ix+iy*lda;
    Bh[i] = Ah[i];
    Bl[i] = Al[i];
  }
}

static __global__
void gddsyrk_add_LL
( int m, 
  double* alphah, double* alphal,
  double Dh[], double Dl[], int ldd,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix>=iy ){
      double workh, workl;
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;

      RMUL( betah, betal, &Ch[ic], &Cl[ic], &Ch[ic], &Cl[ic] );
      RMUL( alphah, alphal, &Dh[id], &Dl[id], &workh, &workl );
      RADD( &workh, &workl, &Ch[ic], &Cl[ic], &Ch[ic], &Cl[ic] );
    }
  }
}

static __global__
void gddsyrk_add_RR
( int m, 
  double* alphah, double* alphal,
  double Dh[], double Dl[], int ldd,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix<=iy ){
      double workh, workl;
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;

      RMUL( betah, betal, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      RMUL( alphah, alphal, Dh+id, Dl+id, &workh, &workl );
      RADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}

static __global__
void gddsyr2k_add_LL
( int m, 
  double* alphah, double* alphal,
  double* Dh, double* Dl, int ldd,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix>=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      int jd = iy+ix*ldc;
      double workh, workl;

      RMUL( betah, betal, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      RADD( Dh+id, Dl+id, Dh+jd, Dl+jd, &workh, &workl );
      RMUL( alphah, alphal, &workh, &workl, &workh, &workl );
      RADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}

static __global__
void gddsyr2k_add_RR
( int m, 
  double* alphah, double* alphal,
  double Dh[], double Dl[], int ldd,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix<=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      int jd = iy+ix*ldc;
      double workh, workl;

      RMUL( betah, betal, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      RADD( Dh+id, Dl+id, Dh+jd, Dl+jd, &workh, &workl );
      RMUL( alphah, alphal, &workh, &workl, &workh, &workl );
      RADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}



extern "C" __global__
void gddgemm_kernel_NN
( int m, int n, int k,
  double* alphah, double* alphal,
  double Ah[], double Al[], int lda,
  double Bh[], double Bl[], int ldb,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 ;
  int ty2 ;
  {
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  tx2 = tx;
  ty2 = ty;
  }
  int tll = tx2;

  __shared__ double Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ double Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  double xxAh[RECTA], xxAl[RECTA];
  double xxBh[RECTB], xxBl[RECTB];

  int trackA = tx2 + ibx + __mul24( ty2, lda );
  int trackB = tx2 + __mul24(iby + ty2*RECTB, ldb );

#pragma unroll
  for(int y=0; y<RECTA; y++){
    Ahs[tx2+ y*TILEWIDTH][ty2] = fetch_x_Ah(trackA + y*TILEWIDTH);
    Als[tx2+ y*TILEWIDTH][ty2] = fetch_x_Al(trackA + y*TILEWIDTH);
  }

#pragma unroll
  for(int y=0; y<RECTB; y++){
    Bhs[tx2][ty2*RECTB+y] = fetch_x_Bh( trackB + y * ldb);
    Bls[tx2][ty2*RECTB+y] = fetch_x_Bl( trackB + y * ldb);
  }

  __syncthreads();

  double Ahxs[RECTA], Alxs[RECTA];
  double Bhxp[RECTB], Blxp[RECTB];
  double Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = 0.0;
      Clb[x*RECTB+y] = 0.0;
    }
  }

  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll += TILEWIDTH;
    trackA += TILEWIDTH*lda ; 
    trackB += TILEWIDTH;

#pragma unroll
    for( int y=0; y<RECTA; y++){
      xxAh[y] = fetch_x_Ah(trackA + y*TILEWIDTH);
      xxAl[y] = fetch_x_Al(trackA + y*TILEWIDTH);
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      xxBh[y] = fetch_x_Bh( trackB + y*ldb);
      xxBl[y] = fetch_x_Bl( trackB + y*ldb);
    }

#pragma unroll 
    for( int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	Ahxs[y] =  Ahs[tx2+y*TILEWIDTH][j1];
	Alxs[y] =  Als[tx2+y*TILEWIDTH][j1];
      }

#pragma unroll
      for( int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2+y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2+y*TILEWIDTH];
      }


#pragma unroll 
      for( int x=0; x<RECTA; x++){
#pragma unroll 
	for( int y=0; y<RECTB; y++){
	  /* C = A*B for each block */
	  double ph, pl;
	  int    cindex = RECTB*x+y;
	  RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y]  += Axs[x]*Bxp[y];
	}
      }
    }

    __syncthreads();
                
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[tx2+y*TILEWIDTH][ty2] = xxAh[y]; 
      Als[tx2+y*TILEWIDTH][ty2] = xxAl[y]; 
    }

#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB + y] = xxBh[y];
      Bls[tx2][ty2*RECTB + y] = xxBl[y];
    }

    __syncthreads();
  }

  Ch += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);
  Cl += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);

#pragma unroll 
  for(int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
      Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
    }
#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhxp[y]= Bhs[j1][ty2 + y*TILEWIDTH];
      Blxp[y]= Bls[j1][ty2 + y*TILEWIDTH];
    }

#pragma unroll 
    for( int x=0; x<RECTA; x++){
#pragma unroll 
      for( int y=0;y<RECTB; y++){
	/* C = A*B for each block */
	double ph, pl;
	int    cindex = RECTB*x+y;
	RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
/*        if(threadIdx.x < m)*/
/*        printf("tid:%d,bid:%d,Ah:%f,Bh:%f,ph:%f\n",threadIdx.x,blockIdx.x,*(Ahxs+x),*(Bhxp+y),ph);*/
	RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	//	Cb[x*RECTB+y] += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby+ty2;
#pragma unroll
  for( int y=0; y<RECTB; y++, gy+=TILEWIDTH ){
    int gx = ibx+tx2; 
#pragma unroll
    for( int x=0; x<RECTA; x++, gx+=TILEWIDTH ){
      if (gx < m && gy < n){
	double ah, al, bh, bl;
	int    cindex = y+x*RECTB;

	  
	RMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	RMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	RADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] = alpha*Cb[y+x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }
    
    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}

extern "C" __global__
void gddgemm_kernel_NT
( int m, int n, int k,
  double* alphah, double* alphal,
  double Ah[], double Al[], int lda,
  double Bh[], double Bl[], int ldb,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 = tx;
  int ty2 = ty;
  int tll = tx2;

  __shared__ double Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ double Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  double xxAh[RECTA], xxAl[RECTA];
  double xxBh[RECTB], xxBl[RECTB];

  int trackA = tx2 + ibx + __mul24(ty2, lda);
  int trackB = tx2 + iby + __mul24(ty2, ldb);        

#pragma unroll
  for(int y=0; y<RECTA; y++){
    Ahs[tx2+ y*TILEWIDTH][ty2] = fetch_x_Ah(trackA + y*TILEWIDTH);
    Als[tx2+ y*TILEWIDTH][ty2] = fetch_x_Al(trackA + y*TILEWIDTH);
  }

#pragma unroll
  for(int y=0; y<RECTB; y++){
    Bhs[ty2][tx2+TILEWIDTH*y] = fetch_x_Bh(trackB+TILEWIDTH*y);
    Bls[ty2][tx2+TILEWIDTH*y] = fetch_x_Bl(trackB+TILEWIDTH*y);
  }

  __syncthreads();

  double Ahxs[RECTA], Alxs[RECTA];
  double Bhxp[RECTB], Blxp[RECTB];
  double Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = 0.0;
      Clb[x*RECTB+y] = 0.0;
    }
  }
        
  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll += TILEWIDTH;
    trackA+=TILEWIDTH*lda ; 
    trackB+=TILEWIDTH*ldb;

#pragma unroll
    for( int y=0; y<RECTA; y++){
      xxAh[y] = fetch_x_Ah(trackA + y*TILEWIDTH);
      xxAl[y] = fetch_x_Al(trackA + y*TILEWIDTH);
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      xxBh[y] = fetch_x_Bh( trackB + TILEWIDTH*y);
      xxBl[y] = fetch_x_Bl( trackB + TILEWIDTH*y);
    }

#pragma unroll 
    for( int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
	Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
      }
#pragma unroll
      for( int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2 + y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2 + y*TILEWIDTH];
      }

#pragma unroll 
      for( int x=0; x<RECTA; x++){
#pragma unroll 
	for( int y=0; y<RECTB; y++){
	  /* C = A*B for each block */
	  double ph, pl;
	  int    cindex = RECTB*x+y;
	  RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y] += Axs[x]*Bxp[y];
	}
      }
    }
    __syncthreads();

#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahs[tx2 + y*TILEWIDTH][ty2] = xxAh[y]; 
      Als[tx2 + y*TILEWIDTH][ty2] = xxAl[y]; 
    }
#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+y*TILEWIDTH] = xxBh[y];
      Bls[ty2][tx2+y*TILEWIDTH] = xxBl[y];
    }

    __syncthreads();
  } 

  Ch += tx2 + ibx + __mul24(ty2 + iby ,ldc);
  Cl += tx2 + ibx + __mul24(ty2 + iby ,ldc);

#pragma unroll 
  for(int j1=0; j1<TILEWIDTH; j1++){
#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
      Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhxp[y] = Bhs[j1][ty2 + y*TILEWIDTH];
      Blxp[y] = Bls[j1][ty2 + y*TILEWIDTH];
    }

#pragma unroll 
    for( int x=0; x<RECTA; x++){
#pragma unroll 
      for( int y=0;y<RECTB;y++){
	/* C = A*B for each block */
	double ph, pl;
	int    cindex = RECTB*x+y;
	RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex,
	      Clb+cindex );
	//	Cb[x*RECTB+y] += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby+ty2;
#pragma unroll
  for( int y=0; y<RECTB; y++, gy+=TILEWIDTH ){
    int gx = ibx+tx2;
#pragma unroll
    for( int x=0; x<RECTA; x++, gx+=TILEWIDTH ){
      if (gx < m && gy < n){
	double ah, al, bh, bl;
	int    cindex = y+x*RECTB;
	  
	RMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	RMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	RADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] = alpha*Cb[y+x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }

    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}

extern "C" __global__
void gddgemm_kernel_TN
( int m, int n, int k,
  double* alphah, double* alphal,
  double Ah[], double Al[], int lda,
  double Bh[], double Bl[], int ldb,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 = tx;
  int ty2 = ty;
  int tll = tx2;

  __shared__ double Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ double Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  double xxAh[RECTA], xxAl[RECTA];
  double xxBh[RECTB], xxBl[RECTB];

  int trackA = tx2 + __mul24( ibx + ty2*RECTA, lda ); 
  int trackB = tx2 + __mul24( iby + ty2*RECTB, ldb ); 

#pragma unroll
  for(int y=0; y<RECTA; y++){
    Ahs[ty2*RECTA+y][tx2] =  (tll<k)*  fetch_x_Ah(trackA + y*lda);
    Als[ty2*RECTA+y][tx2] =  (tll<k)*  fetch_x_Al(trackA + y*lda);
  }

#pragma unroll
  for(int y=0; y<RECTB; y++){
    Bhs[tx2][ty2*RECTB+y] = fetch_x_Bh( trackB + y*ldb );
    Bls[tx2][ty2*RECTB+y] = fetch_x_Bl( trackB + y*ldb );
  }

  __syncthreads();
   
  double Ahxs[RECTA], Alxs[RECTA];
  double Bhxp[RECTB], Blxp[RECTB];
  double Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = 0.0;
      Clb[x*RECTB+y] = 0.0;
    }
  }

  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll +=TILEWIDTH;
    trackA+=TILEWIDTH; 
    trackB+=TILEWIDTH;

#pragma unroll
    for(int y=0; y<RECTA; y++){
      xxAh[y] =  (tll<k)*  fetch_x_Ah(trackA + y*lda);
      xxAl[y] =  (tll<k)*  fetch_x_Al(trackA + y*lda);
    }

#pragma unroll
    for(int y=0; y<RECTB; y++){
      xxBh[y] = fetch_x_Bh(trackB + y*ldb);
      xxBl[y] = fetch_x_Bl(trackB + y*ldb);
    }
#pragma unroll 
    for(int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for(int y=0; y<RECTA; y++){
	Ahxs[y] = Ahs[tx2+y*TILEWIDTH][j1];
	Alxs[y] = Als[tx2+y*TILEWIDTH][j1];
      }
#pragma unroll
      for(int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2+y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2+y*TILEWIDTH];
      }
#pragma unroll 
      for(int x=0; x<RECTA; x++){
#pragma unroll 
	for(int y=0; y<RECTB; y++){
	  /* C = A*B for each block */
	  double ph, pl;
	  int    cindex = RECTB*x+y;
	  RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y]  += Axs[x]*Bxp[y];
	}
      }
    }
    __syncthreads();

#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[ty2*RECTA+y][tx2] = xxAh[y];
      Als[ty2*RECTA+y][tx2] = xxAl[y];
    }

#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB+y] =xxBh[y];
      Bls[tx2][ty2*RECTB+y] =xxBl[y];
    }
    __syncthreads();
  }

  Ch += tx2 + ibx  + __mul24 (ty2 + iby ,ldc);
  Cl += tx2 + ibx  + __mul24 (ty2 + iby ,ldc);

#pragma unroll 
  for(int j1=0; j1<TILEWIDTH; j1++){
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahxs[y] = Ahs[tx2+y*TILEWIDTH][j1];
      Alxs[y] = Als[tx2+y*TILEWIDTH][j1];
    }
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhxp[y]= Bhs[j1][ty2+y*TILEWIDTH];
      Blxp[y]= Bls[j1][ty2+y*TILEWIDTH];
    }
#pragma unroll 
    for(int x=0; x<RECTA; x++){
#pragma unroll 
      for(int y=0; y<RECTB; y++){
	/* C = A*B for each block */
	double ph, pl;
	int    cindex = RECTB*x+y;
	RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	//	Cb[x*RECTB+y] += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby+ty2;
#pragma unroll
  for( int y=0; y<RECTB; y++, gy+=TILEWIDTH ){
    int gx = ibx+tx2;
#pragma unroll
    for( int x=0; x<RECTA; x++, gx+=TILEWIDTH ){
      if (gx < m && gy < n){
	double ah, al, bh, bl;
	int    cindex = y+x*RECTB;
	  
	RMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	RMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	RADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] = alpha*Cb[y+x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }

    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}

extern "C" __global__
void gddgemm_kernel_TT
( int m, int n, int k,
  double* alphah, double* alphal,
  double Ah[], double Al[], int lda,
  double Bh[], double Bl[], int ldb,
  double* betah, double* betal,
  double Ch[], double Cl[], int ldc )
{
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 = tx;
  int ty2 = ty;
  int tll = tx2;

  __shared__ double Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ double Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  double xxAh[RECTA], xxAl[RECTA];
  double xxBh[RECTB], xxBl[RECTB];

  int trackA = tx2 + __mul24( ibx + ty2, lda);
  int trackB = tx2 + iby + __mul24(ty2, ldb);

#pragma unroll
  for(int y=0; y<RECTA; y++){
    Ahs[ty2+TILEWIDTH*y][tx2] = fetch_x_Ah(trackA +  lda*TILEWIDTH*y);
    Als[ty2+TILEWIDTH*y][tx2] = fetch_x_Al(trackA +  lda*TILEWIDTH*y);
  }

#pragma unroll
  for(int y=0; y<RECTB; y++){
    Bhs[ty2][tx2+TILEWIDTH*y] = fetch_x_Bh(trackB+TILEWIDTH*y);
    Bls[ty2][tx2+TILEWIDTH*y] = fetch_x_Bl(trackB+TILEWIDTH*y);
  }

  __syncthreads();

  double Ahxs[RECTA], Alxs[RECTA];
  double Bhxp[RECTB], Blxp[RECTB];
  double Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = 0.0;
      Clb[x*RECTB+y] = 0.0;
    }
  }
        
  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll+=TILEWIDTH;
    trackA+=TILEWIDTH; 
    trackB+=TILEWIDTH*ldb;
                
#pragma unroll
    for( int y=0; y<RECTA; y++){
      xxAh[y] = fetch_x_Ah(trackA + lda*y*TILEWIDTH);
      xxAl[y] = fetch_x_Al(trackA + lda*y*TILEWIDTH);
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      xxBh[y] = fetch_x_Bh(trackB + TILEWIDTH*y);
      xxBl[y] = fetch_x_Bl(trackB + TILEWIDTH*y);
    }

#pragma unroll 
    for( int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
	Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
      }

#pragma unroll
      for( int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2 + y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2 + y*TILEWIDTH];
      }

#pragma unroll 
      for( int x=0; x<RECTA; x++){
#pragma unroll 
	for( int y=0;y<RECTB;y++){
	  /* C = A*B for each block */
	  double ph, pl;
	  int    cindex = RECTB*x+y;
	  RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y] += Axs[x]*Bxp[y];
	}
      }
    }
    __syncthreads();

#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahs[ty2 + TILEWIDTH*y][tx2] = xxAh[y];
      Als[ty2 + TILEWIDTH*y][tx2] = xxAl[y];
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+y*TILEWIDTH] = xxBh[y];
      Bls[ty2][tx2+y*TILEWIDTH] = xxBl[y];
    }

    __syncthreads();
  } 

  Ch += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);
  Cl += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);

#pragma unroll 
  for( int j1=0; j1<TILEWIDTH; j1++){
#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahxs[y] =  Ahs[tx2 + y*TILEWIDTH][j1];
      Alxs[y] =  Als[tx2 + y*TILEWIDTH][j1];
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhxp[y] = Bhs[j1][ty2 + y*TILEWIDTH];
      Blxp[y] = Bls[j1][ty2 + y*TILEWIDTH];
    }

#pragma unroll 
    for( int x=0; x<RECTA; x++){
#pragma unroll 
      for( int y=0; y<RECTB; y++){
	/* C = A*B for each block */
	double ph, pl;
	int    cindex = RECTB*x+y;
	RMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	RADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex,
	      Clb+cindex );
	//	Cb[x*RECTB+y] += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby+ty2;
#pragma unroll
  for( int y=0; y<RECTB; y++, gy+=TILEWIDTH ){
    int gx = ibx+tx2; 
#pragma unroll
    for( int x=0; x<RECTA; x++, gx+=TILEWIDTH ){
      if (gx < m && gy < n){
	double ah, al, bh, bl;
	int    cindex = y+x*RECTB;
	  
	RMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	RMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	RADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] = alpha*Cb[y+x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }

    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}



__global__  void
print_kernel(double *Ah,char c)
{
int tx = threadIdx.x;
int ty = threadIdx.y;
int lda = blockDim.x;
printf("%c::tx:%d,ty:%d:%e\n",c,tx,ty,*(Ah+tx+lda*ty));
}
//----------------------------------------------
extern "C" void gddgemm
( char* transA, char* transB,
  int* m, int* n, int* k,
  double* alphah, double* alphal,
  double Ah[], double Al[], int* lda,
  double Bh[], double Bl[], int *ldb,
  double* betah, double* betal,
  double Ch[], double Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;

  if( *transA == 'N' || *transA == 'n' ){
    if( *transB == 'N' || *transB == 'n' ){
      // C(M,N) += A(M,K)*B(K,N), A and B are general
      nrowA = *m, ncolA = *k;
      nrowB = *k, ncolB = *n;
      nrowC = *m, ncolC = *n;
    }
    else{
      // C(M,N) += A(M,K)*B(N,K)^T, A and B are general
      nrowA = *m, ncolA = *k;
      nrowB = *n, ncolB = *k;
      nrowC = *m, ncolC = *n;
    }
  }
  else{
    if( *transB == 'N' || *transB == 'n' ){
      // A(K,M)^T *B(K,N), A and B are general
      nrowA = *k, ncolA = *m;
      nrowB = *k, ncolB = *n;
      nrowC = *m, ncolC = *n;
    }
    else{
      // A(K,M)^T *B(K,N)^T, A and B are general
      nrowA = *k, ncolA = *m;
      nrowB = *n, ncolB = *k;
      nrowC = *m, ncolC = *n;
    }
  }
/*  dim3 ba(nrowA,ncolA);*/
/*  print_kernel<<<1,ba>>>*/
/*          (Ah,'A');*/
 
/*  dim3 bb(nrowB,ncolB);*/
/*  print_kernel<<<1,bb>>>*/
/*          (Bh,'B');*/
  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddgemm\n");
    printf("lda=%d.nrowA=%d,ldb=%d,nrowB=%d,ldc=%d,nrowC=%d\n",*lda,nrowA,*ldb,nrowB,*ldc,nrowC);
    printf("wrong lda or ldb or ldc.\n");
    exit(-1);
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;

  // bind data of matrix A and B to texture memory
  if(
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah, sizeA*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Al, (int2*)Al, sizeA*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bh, (int2*)Bh, sizeB*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bl, (int2*)Bl, sizeB*sizeof(double) ) ){
    printf( "cannot bind to texture in gddgemm\n" );
    return;
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*m)/BLK_A+((*m)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    if( *transA == 'N' || *transA == 'n' ){
      if( *transB == 'N' || *transB == 'n' ){
	// C += A*B
	gddgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	  ( *m,  *n,  *k,  alphah, alphal, Ah,  Al,  *lda, Bh,  Bl,  *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
      else{
	// C += A*B^T
	gddgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	  ( *m,  *n,  *k,  alphah, alphal, Ah,  Al,  *lda, Bh,  Bl,  *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
    }
    else{
      if( *transB == 'N' || *transB == 'n' ){
	// C += A^T*B
	gddgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	  ( *m,  *n,  *k,  alphah, alphal, Ah,  Al,  *lda, Bh,  Bl,  *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
      else{
	// C += A^T*B^T
	gddgemm_kernel_TT<<< blocks, threads, 0, gddstream>>>
	  ( *m,  *n,  *k,  alphah, alphal, Ah,  Al,  *lda, Bh,  Bl,  *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
    }


/*    printf("nrowC:%d,ncolC%d\n",nrowC,ncolC);*/
/*  dim3 bc(nrowC,ncolC);*/
/*  print_kernel<<<1,bc>>>*/
/*          (Ch,'C');*/

  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );
}

//----------------------------------------------
extern "C" void gddsymm
( char* side, char* uplo,
  int* m, int* n,
  double* alphah, double* alphal,
  double Ah[], double Al[], int* lda,
  double Bh[], double Bl[], int* ldb,
  double* betah, double* betal,
  double Ch[], double Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;

  if( *side == 'L' || *side == 'l' ){
    // C(M,N) += A(M,M)*B(M,N), A is symmetric, B is general
    nrowA = *m, ncolA = *m;
    nrowB = *m, ncolB = *n;
    nrowC = *m, ncolC = *n;
  }
  else{
    // C(M,N) += B(M,N)*A(N,N), A is symmetric, B is general
    nrowB = *m, ncolB = *n;
    nrowA = *n, ncolA = *n;
    nrowC = *m, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddsymm\n");
    printf("wrong lda or ldb or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;

  // construct full matrix from triangular/symmetrix matrix
  double *Afullh, *Afulll;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&Afullh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&Afulll ) ){
    cublasFree( Afullh );
    cublasFree( Afulll );
    printf("Device Malloc Error in gddsymm\n");
    return;
  }

  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( nrowA/TILEWIDTH+(nrowA%TILEWIDTH!=0),
		 ncolA/TILEWIDTH+(ncolA%TILEWIDTH!=0) );

    if( *uplo == 'L' || *uplo == 'l' ){
      gddSYLtoFull<<< blocks, threads, 0, gddstream>>>
	( ncolA, Ah, Al, Afullh, Afulll, *lda );
    }
    else{
      gddSYUtoFull<<< blocks, threads, 0, gddstream>>>
	( ncolA, Ah, Al, Afullh, Afulll, *lda );
    }
  }

  if( *side == 'L' || *side == 'l' ){
    // bind data of matrix A and B to texture memory
    if(
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Ah, (int2*)Afullh, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Al, (int2*)Afulll, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bh, (int2*)Bh, sizeB*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bl, (int2*)Bl, sizeB*sizeof(double) ) ){
      printf( "cannot bind to texture in gddsymm\n" );
      return;
    }
  }
  else{
    // bind data of matrix B and A to texture memory
    if(
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bh, (int2*)Afullh, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bl, (int2*)Afulll, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Ah, (int2*)Bh, sizeB*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Al, (int2*)Bl, sizeB*sizeof(double) ) ){
      printf( "cannot bind to texture in gddsymm\n" );
      return;
    }
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*m)/BLK_A+((*m)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    if( *side == 'L' || *side == 'l' ){
      // C += A*B
      gddgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	( *m, *n, *m,
	  alphah, alphal,
	  Afullh, Afulll, *lda,
	  Bh, Bl, *ldb,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
    else{
      // C += B*A
      gddgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	( *m, *n, *n,
	  alphah, alphal,
	  Bh, Bl, *ldb,
	  Afullh, Afulll, *lda,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Afullh );
  cublasFree( Afulll );
}

//----------------------------------------------
extern "C" void gddtrmm
( char* side, char* uplo, char* transA, char* diag,
  int* m, int* n,
  double* alphah, double* alphal,
  double Ah[], double Al[], int* lda,
  double Bh[], double Bl[], int* ldb )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;

  if( *side == 'L' || *side == 'l' ){
    // B(M,N) = A(M,M)*B(M,N), A is triangular, B is general
    nrowA = *m, ncolA = *m;
    nrowB = *m, ncolB = *n;
  }
  else{
    // B(M,N) += B(M,N)*A(N,N), A is triangular, B is general
    nrowB = *m, ncolB = *n;
    nrowA = *n, ncolA = *n;
  }

  if( *lda < nrowA || *ldb < nrowB ){
    printf("Argument Error in gddtrmm\n");
    printf("wrong lda or ldb.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;

  // construct full matrix from triangular/symmetrix matrix
  double *Afullh, *Afulll;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&Afullh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&Afulll ) ){
    cublasFree( Afullh );
    cublasFree( Afulll );
    printf("Device Malloc Error in gddtrmm\n");
    return;
  }

  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( nrowA/TILEWIDTH+(nrowA%TILEWIDTH!=0),
		 ncolA/TILEWIDTH+(ncolA%TILEWIDTH!=0) );

    if( *uplo == 'L' || *uplo == 'l' ){
      if( *diag == 'U' || *diag == 'u' ){
	gddTLDtoFull<<< blocks, threads, 0, gddstream>>>
	  ( ncolA, Ah, Al, Afullh, Afulll, *lda );
      }
      else{
	gddTLtoFull <<< blocks, threads, 0, gddstream>>>
	  ( ncolA, Ah, Al, Afullh, Afulll, *lda );
      }
    }
    else{
      if( *diag == 'U' || *diag == 'u' ){
	gddTUDtoFull<<< blocks, threads, 0, gddstream>>>
	  ( ncolA, Ah, Al, Afullh, Afulll, *lda );
      }
      else{
	gddTUtoFull <<< blocks, threads, 0, gddstream>>>
	  ( ncolA, Ah, Al, Afullh, Afulll, *lda );
      }
    }
  }

  // allocate temporal arrays
  double *Ch, *Cl;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&Ch ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&Cl ) ){
    cublasFree( Ch );
    cublasFree( Cl );
    printf("Device Malloc Error in gddtrmm\n");
    return;
  }

  if( *side == 'L' || *side == 'l' ){
    // bind data of matrix A and B to texture memory
    if(
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Ah, (int2*)Afullh, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Al, (int2*)Afulll, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bh, (int2*)Bh, sizeB*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bl, (int2*)Bl, sizeB*sizeof(double) ) ){
      printf( "cannot bind to texture in gddtrmm\n" );
      return;
    }
  }
  else{
    // bind data of matrix B and A to texture memory
    if(
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bh, (int2*)Afullh, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Bl, (int2*)Afulll, sizeA*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Ah, (int2*)Bh, sizeB*sizeof(double) ) ||
       cudaSuccess !=
       cudaBindTexture( NULL, texture_Al, (int2*)Bl, sizeB*sizeof(double) ) ){
      printf( "cannot bind to texture in gddtrmm\n" );
      return;
    }
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*m)/BLK_A+((*m)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );
    
    // make a zero on devive
    double zerohh = 0.0, zerohl = 0.0;
    double *zeroh, *zerol;
    cublasAlloc( 1, sizeof(double), (void**)&zeroh );
    cublasAlloc( 1, sizeof(double), (void**)&zerol );
    cublasSetVector( 1, sizeof(double), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(double), &zerohl, 1, zerol, 1 );

    if( *side == 'L' || *side == 'l' ){
      if( *transA == 'N' || *transA == 'n' ){
	// C = alpha*A*B
	gddgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *m,
	    alphah, alphal,
	    Afullh, Afulll, *lda,
	    Bh, Bl, *ldb,
	    zeroh, zerol,
	    Ch, Cl, *ldb );
      }
      else{
	// C = alpha*A^T*B
	gddgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *m,
	    alphah, alphal,
	    Afullh, Afulll, *lda,
	    Bh, Bl, *ldb,
	    zeroh, zerol,
	    Ch, Cl, *ldb );
      }
    }
    else{
      if( *transA == 'N' || *transA == 'n' ){
	// C = alpha*B*A
	gddgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *n,
	    alphah, alphal,
	    Bh, Bl, *ldb,
	    Afullh, Afulll, *lda,
	    zeroh, zerol,
	    Ch, Cl, *ldb );
      }
      else{
	// C = alpha*B*A^T
	gddgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *n,
	    alphah, alphal,
	    Bh, Bl, *ldb,
	    Afullh, Afulll, *lda,
	    zeroh, zerol,
	    Ch, Cl, *ldb );
      }
    }
    cublasFree( zeroh );
    cublasFree( zerol );
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  // update B by C
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( nrowB/TILEWIDTH+(nrowB%TILEWIDTH!=0),
		 ncolB/TILEWIDTH+(ncolB%TILEWIDTH!=0) );

    gddFulltoFull<<< blocks, threads, 0, gddstream>>>
      (nrowB, ncolB, Ch, Ch, Bh, Bl, *ldb );
  }

  cublasFree( Afullh );
  cublasFree( Afulll );
  cublasFree( Ch );
  cublasFree( Cl );
}

//----------------------------------------------
extern "C" void gddsyr2k
( char* uplo, char* trans,
  int* n, int* k,
  double* alphah, double* alphal,
  double Ah[], double Al[], int* lda,
  double Bh[], double Bl[], int *ldb,
  double* betah, double* betal,
  double Ch[], double Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;

  if( *trans == 'N' || *trans == 'n' ){
    // C(N,N) += A(N,K) * B(N,K)^T
    nrowA = *n, ncolA = *k;
    nrowB = *n, ncolB = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    // C(N,N) += A(K,N)^T * B(K,N)
    nrowA = *k, ncolA = *n;
    nrowB = *k, ncolB = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddsyr2k\n");
    printf("wrong lda or ldb or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;
  int sizeC = (*ldc)*ncolC;

  // bind data of matrix A and B to texture memory
  if(
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah, sizeA*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Al, (int2*)Al, sizeA*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bh, (int2*)Bh, sizeB*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bl, (int2*)Bl, sizeB*sizeof(double) ) ){
    printf( "cannot bind to texture in gddsyr2k\n" );
    return;
  }

  // allocate temporal arrays
  double *Dh, *Dl;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**)&Dh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**)&Dl ) ){
    cublasFree( Dh );
    cublasFree( Dl );
    printf("Device Malloc Error in gddsyr2k\n");
    return;
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/BLK_A+((*n)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    // make a one on devive
    double onehh = 1.0, onehl = 0.0;
    double *oneh, *onel;
    cublasAlloc( 1, sizeof(double), (void**)&oneh );
    cublasAlloc( 1, sizeof(double), (void**)&onel );
    cublasSetVector( 1, sizeof(double), &onehh, 1, oneh, 1 );
    cublasSetVector( 1, sizeof(double), &onehl, 1, onel, 1 );

    // make a zero on devive
    double zerohh = 0.0, zerohl = 0.0;
    double *zeroh, *zerol;
    cublasAlloc( 1, sizeof(double), (void**)&zeroh );
    cublasAlloc( 1, sizeof(double), (void**)&zerol );
    cublasSetVector( 1, sizeof(double), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(double), &zerohl, 1, zerol, 1 );

    if( *trans == 'N' || *trans == 'n' ){
      // D = 1.0*A*B^T + 0.0*D
      gddgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Bh,  Bl, *ldb,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
    else{
      // D = 1.0*alpha*A^T*B + 0.0*D
      gddgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Bh,  Bl, *ldb,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }

    cublasFree( zeroh );
    cublasFree( zerol );
    cublasFree( oneh );
    cublasFree( onel );
  }

  // exec extra kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0),
		 (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0) );

    // C = alpha*(D + D^T) + beta*C
    if( *uplo == 'L' || *uplo == 'l' ){
      gddsyr2k_add_LL <<< blocks, threads, 0, gddstream>>>
	( *n,
	  alphah, alphal,
	  Dh, Dl, *ldc,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
    else{
      gddsyr2k_add_RR <<< blocks, threads, 0, gddstream>>>
	( *n,
	  alphah, alphal,
	  Dh, Dl, *ldc,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Dh );
  cublasFree( Dl );
}


//----------------------------------------------
extern "C" void gddsyrk
( char* uplo, char* trans,
  int* n, int* k,
  double* alphah, double* alphal,
  double Ah[], double Al[], int* lda,
  double* betah, double* betal,
  double Ch[], double Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowC, ncolC;

  if( *trans == 'N' || *trans == 'n' ){
    // C(N,N) += A(N,K) * A(N,K)^T
    nrowA = *n, ncolA = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    // C(N,N) += A(K,N)^T * A(K,N)
    nrowA = *k, ncolA = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldc < nrowC ){
    printf("Argument Error in gddsyrk\n");
    printf("wrong lda or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeC = (*ldc)*ncolC;

  // bind data of matrix A and B to texture memory
  if(
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah, sizeA*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Al, (int2*)Al, sizeA*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bh, (int2*)Ah, sizeA*sizeof(double) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bl, (int2*)Al, sizeA*sizeof(double) ) ){
    printf( "cannot bind to texture in gddsyrk\n" );
    return;
  }

  // allocate temporal arrays
  double *Dh, *Dl;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**)&Dh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**)&Dl ) ){
    cublasFree( Dh );
    cublasFree( Dl );
    printf("Device Malloc Error in gddsyrk\n");
    return;
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/BLK_A+((*n)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    // make a one on devive
    double onehh = 1.0, onehl = 0.0;
    double *oneh, *onel;
    cublasAlloc( 1, sizeof(double), (void**)&oneh );
    cublasAlloc( 1, sizeof(double), (void**)&onel );
    cublasSetVector( 1, sizeof(double), &onehh, 1, oneh, 1 );
    cublasSetVector( 1, sizeof(double), &onehl, 1, onel, 1 );

    // make a zero on devive
    double zerohh = 0.0, zerohl = 0.0;
    double *zeroh, *zerol;
    cublasAlloc( 1, sizeof(double), (void**)&zeroh );
    cublasAlloc( 1, sizeof(double), (void**)&zerol );
    cublasSetVector( 1, sizeof(double), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(double), &zerohl, 1, zerol, 1 );

    // D = 1.0*A*A^T + 0.0*D
    if( *trans == 'N' || *trans == 'n' ){
      gddgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Ah, Al, *lda,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }

    // D = 1.0*alpha*A^T*A + 0.0*D
    else{
      gddgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Ah, Al, *lda,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
  }

  // exec extra kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0),
		 (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0) );

    // C = alpha*(D) + beta*C
    if( *uplo == 'L' || *uplo == 'l' ){
      gddsyrk_add_LL <<< blocks, threads, 0, gddstream>>>
	( *n, alphah, alphal, Dh, Dl, *ldc, betah, betal, Ch, Cl, *ldc );
    }
    else{
      gddsyrk_add_RR <<< blocks, threads, 0, gddstream>>>
	( *n, alphah, alphal, Dh, Dl, *ldc, betah, betal, Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Dh );
  cublasFree( Dl );
}
