#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

#define TILEWIDTH ( 16 )

#define RECTA ( 2 )
#define RECTB ( 2 )

#define BLK_A ( TILEWIDTH*RECTA )
#define BLK_B ( TILEWIDTH*RECTB )

texture<int2, 1> texture_Ah, texture_Al;
texture<int2, 1> texture_Bh, texture_Bl;

static __inline__ __device__
cuDoubleComplex fetch_x_Ah( const int& i )
{
  register int2 vr = tex1Dfetch( texture_Ah, 2*i+0 );
  register int2 vi = tex1Dfetch( texture_Ah, 2*i+1 );
  
  return make_cuDoubleComplex
    ( __hiloint2double( vr.y, vr.x ), __hiloint2double( vi.y, vi.x ) );
}

static __inline__ __device__
cuDoubleComplex fetch_x_Al( const int& i )
{
  register int2 vr = tex1Dfetch( texture_Al, 2*i+0 );
  register int2 vi = tex1Dfetch( texture_Al, 2*i+1 );
  
  return make_cuDoubleComplex
    ( __hiloint2double( vr.y, vr.x ), __hiloint2double( vi.y, vi.x ) );
}

static __inline__ __device__
cuDoubleComplex fetch_x_Bh( const int &i )
{
  register int2 vr = tex1Dfetch( texture_Bh, 2*i+0 );
  register int2 vi = tex1Dfetch( texture_Bh, 2*i+1 );

  return make_cuDoubleComplex
    ( __hiloint2double( vr.y, vr.x ), __hiloint2double( vi.y, vi.x ) );
}

static __inline__ __device__
cuDoubleComplex fetch_x_Bl( const int &i )
{
  register int2 vr = tex1Dfetch( texture_Bl, 2*i+0 );
  register int2 vi = tex1Dfetch( texture_Bl, 2*i+1 );

  return make_cuDoubleComplex
    ( __hiloint2double( vr.y, vr.x ), __hiloint2double( vi.y, vi.x ) );
}


// construct a full matrix from lower part of symmetric matrix
static __global__
void gddzHELtoFull
( int m,
  cuDoubleComplex Atrih[], cuDoubleComplex Atril[],
  cuDoubleComplex Afullh[], cuDoubleComplex Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix==iy ){
      int i = ix+iy*lda;
      Afullh[i] = make_cuDoubleComplex( cuCreal(Atrih[i]), 0.0 );
      Afulll[i] = make_cuDoubleComplex( cuCreal(Atril[i]), 0.0 );
    }
    else
    if( ix>iy ){
      int i = ix+iy*lda;
      Afullh[i] = Atrih[i];
      Afulll[i] = Atril[i];
    }
    else{
      int i = ix+iy*lda;
      int j = iy+ix*lda;
      Afullh[i] = cuConj(Atrih[j]);
      Afulll[i] = cuConj(Atril[j]);
    }
  }
}

// construct a full matrix from upper part of symmetric matrix
static __global__
void gddzHEUtoFull
( int m,
  cuDoubleComplex Atrih[], cuDoubleComplex Atril[],
  cuDoubleComplex Afullh[], cuDoubleComplex Afulll[], int lda )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix==iy ){
      int i = ix+iy*lda;
      Afullh[i] = make_cuDoubleComplex( cuCreal(Atrih[i]), 0.0 );
      Afulll[i] = make_cuDoubleComplex( cuCreal(Atril[i]), 0.0 );
    }
    else if( ix<iy ){
      int i = ix+iy*lda;
      Afullh[i] = Atrih[i];
      Afulll[i] = Atril[i];
    }
    else{
      int i = ix+iy*lda;
      int j = iy+ix*lda;
      Afullh[i] = cuConj(Atrih[j]);
      Afulll[i] = cuConj(Atril[j]);
    }
  }
}

static __global__
void gddzsyrk_add_LL
( int m, 
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix>=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      cuDoubleComplex workh, workl;

      CMUL( betah, betal, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      CMUL( alphah, alphal, Dh+id, Dl+id, &workh, &workl );
      CADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}

static __global__
void gddzsyrk_add_RR
( int m, 
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix<=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      cuDoubleComplex workh, workl;

      CMUL( betah, betal, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      CMUL( alphah, alphal, Dh+id, Dl+id, &workh, &workl );
      CADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}

static __global__
void gddzherk_add_LL
( int m, 
 double* alphah, double* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  double* betah, double* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix>=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;

      if(ix == iy)
      {
      *(Ch+ic) = make_cuDoubleComplex(Ch[ic].x,0.0);
      *(Cl+ic) = make_cuDoubleComplex(Cl[ic].x,0.0);
      }

      cuDoubleComplex workh, workl;

      workh = make_cuDoubleComplex(*betah,0.0);
      workl = make_cuDoubleComplex(*betal,0.0);
      CMUL( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      workh = make_cuDoubleComplex(*alphah,0.0);
      workl = make_cuDoubleComplex(*alphal,0.0);
      CMUL( &workh, &workl, Dh+id, Dl+id, &workh, &workl );
      CADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}

static __global__
void gddzherk_add_RR
( int m, 
  double* alphah, double* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  double* betah, double* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix<=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;


  if(ix == iy)
      {
      *(Ch+ic) = make_cuDoubleComplex(Ch[ic].x,0.0);
      *(Cl+ic) = make_cuDoubleComplex(Cl[ic].x,0.0);
      }


      cuDoubleComplex workh, workl;

      workh = make_cuDoubleComplex(*betah,0.0);
      workl = make_cuDoubleComplex(*betal,0.0);
      CMUL( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      workh = make_cuDoubleComplex(*alphah,0.0);
      workl = make_cuDoubleComplex(*alphal,0.0);
      CMUL( &workh, &workl, Dh+id, Dl+id, &workh, &workl );
      CADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}


static __global__
void gddzsyr2k_add_LL
( int m, 
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix>=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      int jd = iy+ix*ldc;
      cuDoubleComplex workh, workl;

      CMUL( betah, betal, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      workh = Dh[jd]; // no conj
      workl = Dl[jd]; // no conj
      CADD( Dh+id, Dl+id, &workh, &workl, &workh, &workl );
      CMUL( alphah, alphal, &workh, &workl, &workh, &workl );
      CADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}

static __global__
void gddzher2k_add_LL
( int m, 
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  double* betah, double* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix>=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      int jd = iy+ix*ldc;
      cuDoubleComplex workh1, workl1;
      cuDoubleComplex workh2, workl2;

      if(ix == iy)
      {
      *(Ch+ic) = make_cuDoubleComplex(Ch[ic].x,0.0);
      *(Cl+ic) = make_cuDoubleComplex(Cl[ic].x,0.0);
      }


      workh1 = make_cuDoubleComplex(*betah,0.0);
      workl1 = make_cuDoubleComplex(*betal,0.0);
      CMUL( &workh1, &workl1, Ch+ic, Cl+ic, Ch+ic, Cl+ic );

      CMUL( alphah, alphal, &Dh[id], &Dl[id], &workh1, &workl1 );
      CMUL( alphah, alphal, &Dh[jd], &Dl[jd], &workh2, &workl2 );
      workh2 = cuConj(workh2);
      workl2 = cuConj(workl2);

      CADD( &workh1, &workl1, &workh2, &workl2, &workh1, &workl1 );
      CADD( &workh1, &workl1, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}



static __global__
void gddzsyr2k_add_RR
( int m, 
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix<=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      int jd = iy+ix*ldc;
      cuDoubleComplex workh, workl;

      CMUL( betah, betal, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
      workh = Dh[jd]; // no conj
      workl = Dl[jd]; // no conj
      CADD( Dh+id, Dl+id, &workh, &workl, &workh, &workl );
      CMUL( alphah, alphal, &workh, &workl, &workh, &workl );
      CADD( &workh, &workl, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}


static __global__
void gddzher2k_add_RR
( int m, 
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Dh[], cuDoubleComplex Dl[], int ldd,
  double* betah, double* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx = threadIdx.x;
  int ty = threadIdx.y;
  int bx = blockIdx.x;
  int by = blockIdx.y;
  int ix = tx+bx*TILEWIDTH;
  int iy = ty+by*TILEWIDTH;

  if( ix<m && iy<m ){
    if( ix<=iy ){
      int ic = ix+iy*ldc;
      int id = ix+iy*ldd;
      int jd = iy+ix*ldc;
      if(ix == iy)
      {
      *(Ch+ic) = make_cuDoubleComplex(Ch[ic].x,0.0);
      *(Cl+ic) = make_cuDoubleComplex(Cl[ic].x,0.0);
      }
      cuDoubleComplex workh1, workl1;
      cuDoubleComplex workh2, workl2;

      workh1 = make_cuDoubleComplex(*betah,0.0);
      workl1 = make_cuDoubleComplex(*betal,0.0);
      CMUL( &workh1, &workl1, Ch+ic, Cl+ic, Ch+ic, Cl+ic );

      CMUL( alphah, alphal, &Dh[id], &Dl[id], &workh1, &workl1 );
      CMUL( alphah, alphal, &Dh[jd], &Dl[jd], &workh2, &workl2 );
      workh2 = cuConj(workh2);
      workl2 = cuConj(workl2);

      CADD( &workh1, &workl1, &workh2, &workl2, &workh1, &workl1 );
      CADD( &workh1, &workl1, Ch+ic, Cl+ic, Ch+ic, Cl+ic );
    }
  }
}





extern "C" __global__
void gddzgemm_kernel_NN
( int m, int n, int k,
  bool conjA, bool conjB,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 = tx;
  int ty2 = ty;
  int tll = tx2;

  __shared__ cuDoubleComplex Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ cuDoubleComplex Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  cuDoubleComplex xxAh[RECTA], xxAl[RECTA];
  cuDoubleComplex xxBh[RECTB], xxBl[RECTB];

  /* coalessed access */
  int trackA = tx2 + ibx + __mul24( ty2, lda );
  int trackB = tx2 + __mul24(iby + ty2*RECTB, ldb );

  if( !conjA ){
#pragma unroll
    for(int x=0; x<RECTA; x++){
      Ahs[tx2+ x*TILEWIDTH][ty2] = fetch_x_Ah(trackA + x*TILEWIDTH);
      Als[tx2+ x*TILEWIDTH][ty2] = fetch_x_Al(trackA + x*TILEWIDTH);
    }
  }
  else{
#pragma unroll
    for(int x=0; x<RECTA; x++){
      Ahs[tx2+ x*TILEWIDTH][ty2] = cuConj(fetch_x_Ah(trackA + x*TILEWIDTH));
      Als[tx2+ x*TILEWIDTH][ty2] = cuConj(fetch_x_Al(trackA + x*TILEWIDTH));
    }
  }

  if( !conjB ){
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB+y] = fetch_x_Bh(trackB + y*ldb);
      Bls[tx2][ty2*RECTB+y] = fetch_x_Bl(trackB + y*ldb);
    }
  }
  else{
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB+y] = cuConj(fetch_x_Bh(trackB + y*ldb));
      Bls[tx2][ty2*RECTB+y] = cuConj(fetch_x_Bl(trackB + y*ldb));
    }
  }

  __syncthreads();

  cuDoubleComplex Ahxs[RECTA], Alxs[RECTA];
  cuDoubleComplex Bhxp[RECTB], Blxp[RECTB];
  cuDoubleComplex Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
      Clb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
    }
  }

  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll += TILEWIDTH;
    trackA += TILEWIDTH*lda ; 
    trackB += TILEWIDTH;

    if( !conjA ){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	xxAh[y] = fetch_x_Ah(trackA + y*TILEWIDTH);
	xxAl[y] = fetch_x_Al(trackA + y*TILEWIDTH);
      }
    }
    else{
#pragma unroll
      for( int y=0; y<RECTA; y++){
	xxAh[y] = cuConj(fetch_x_Ah(trackA + y*TILEWIDTH));
	xxAl[y] = cuConj(fetch_x_Al(trackA + y*TILEWIDTH));
      }
    }

    if( !conjB ){
#pragma unroll
      for( int y=0; y<RECTB; y++){
	xxBh[y] = fetch_x_Bh( trackB + y*ldb);
	xxBl[y] = fetch_x_Bl( trackB + y*ldb);
      }
    }
    else{
#pragma unroll
      for( int y=0; y<RECTB; y++){
	xxBh[y] = cuConj(fetch_x_Bh( trackB + y*ldb));
	xxBl[y] = cuConj(fetch_x_Bl( trackB + y*ldb));
      }
    }

#pragma unroll 
    for( int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	Ahxs[y] = Ahs[tx2+y*TILEWIDTH][j1];
	Alxs[y] = Als[tx2+y*TILEWIDTH][j1];
      }

#pragma unroll
      for( int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2+y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2+y*TILEWIDTH];
      }


#pragma unroll 
      for( int x=0; x<RECTA; x++){
#pragma unroll 
	for( int y=0; y<RECTB; y++){
	  /* C = A*B for each block */
	  cuDoubleComplex ph, pl;
	  int    cindex = RECTB*x+y;
	  CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y]  += Axs[x]*Bxp[y];
	}
      }
    }

    __syncthreads();
                
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[tx2+y*TILEWIDTH][ty2] = xxAh[y]; 
      Als[tx2+y*TILEWIDTH][ty2] = xxAl[y]; 
    }

#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB + y] = xxBh[y];
      Bls[tx2][ty2*RECTB + y] = xxBl[y];
    }

    __syncthreads();
  }

  Ch += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);
  Cl += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);

#pragma unroll 
  for(int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
      Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
    }
#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhxp[y]= Bhs[j1][ty2 + y*TILEWIDTH];
      Blxp[y]= Bls[j1][ty2 + y*TILEWIDTH];
    }

#pragma unroll 
    for( int x=0; x<RECTA; x++){
#pragma unroll 
      for( int y=0;y<RECTB; y++){
	/* C = A*B for each block */
	cuDoubleComplex ph, pl;
	int    cindex = RECTB*x+y;
	CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex,
	      Clb+cindex );
	//	Cb[x*RECTB + y]  += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby + ty2;
#pragma unroll
  for( int y=0;y<RECTB;y++, gy+=TILEWIDTH){
    int gx = ibx + tx2; 
#pragma unroll
    for(int x=0;x<RECTA;x++, gx+=TILEWIDTH){
      if (gx < m && gy < n){
	cuDoubleComplex ah, al, bh, bl;
	int    cindex = y+x*RECTB;
	  
	CMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	CMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	CADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] = alpha*Cb[y+x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }
    
    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}

extern "C" __global__
void gddzgemm_kernel_NT
( int m, int n, int k,
  bool conjA, bool conjB,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 = tx;
  int ty2 = ty;
  int tll = tx2;

  __shared__ cuDoubleComplex Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ cuDoubleComplex Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  cuDoubleComplex xxAh[RECTA], xxAl[RECTA];
  cuDoubleComplex xxBh[RECTB], xxBl[RECTB];

  int trackA = tx2 + ibx +__mul24(ty2, lda);
  int trackB = tx2 + iby + __mul24(ty2, ldb);        

  if( !conjA ){
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[tx2+ y*TILEWIDTH][ty2] = fetch_x_Ah(trackA + y*TILEWIDTH);
      Als[tx2+ y*TILEWIDTH][ty2] = fetch_x_Al(trackA + y*TILEWIDTH);
    }
  }
  else{
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[tx2+ y*TILEWIDTH][ty2] = cuConj(fetch_x_Ah(trackA + y*TILEWIDTH));
      Als[tx2+ y*TILEWIDTH][ty2] = cuConj(fetch_x_Al(trackA + y*TILEWIDTH));
    }
  }

  if( !conjB ){
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+TILEWIDTH*y] = fetch_x_Bh(trackB+TILEWIDTH*y);
      Bls[ty2][tx2+TILEWIDTH*y] = fetch_x_Bl(trackB+TILEWIDTH*y);
    }
  }
  else{
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+TILEWIDTH*y] = cuConj(fetch_x_Bh(trackB+TILEWIDTH*y));
      Bls[ty2][tx2+TILEWIDTH*y] = cuConj(fetch_x_Bl(trackB+TILEWIDTH*y));
    }
  }
  __syncthreads();

  cuDoubleComplex Ahxs[RECTA], Alxs[RECTA];
  cuDoubleComplex Bhxp[RECTB], Blxp[RECTB];
  cuDoubleComplex Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
    }
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Clb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
    }
  }
        
  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll += TILEWIDTH;
    trackA+=TILEWIDTH*lda ; 
    trackB+=TILEWIDTH*ldb;

    if( !conjA ){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	xxAh[y] = fetch_x_Ah(trackA + y*TILEWIDTH);
	xxAl[y] = fetch_x_Al(trackA + y*TILEWIDTH);
      }
    }
    else{
#pragma unroll
      for( int y=0; y<RECTA; y++){
	xxAh[y] = cuConj(fetch_x_Ah(trackA + y*TILEWIDTH));
	xxAl[y] = cuConj(fetch_x_Al(trackA + y*TILEWIDTH));
      }
    }

    if( !conjB ){
#pragma unroll
      for( int y=0; y<RECTB; y++){
	xxBh[y] = fetch_x_Bh( trackB + TILEWIDTH*y);
	xxBl[y] = fetch_x_Bl( trackB + TILEWIDTH*y);
      }
    }
    else{
#pragma unroll
      for( int y=0; y<RECTB; y++){
	xxBh[y] = cuConj(fetch_x_Bh( trackB + TILEWIDTH*y));
	xxBl[y] = cuConj(fetch_x_Bl( trackB + TILEWIDTH*y));
      }
    }

#pragma unroll 
    for( int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
	Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
      }
#pragma unroll
      for( int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2 + y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2 + y*TILEWIDTH];
      }

#pragma unroll 
      for( int x=0; x<RECTA; x++){
#pragma unroll 
	for( int y=0; y<RECTB; y++){
	  /* C = A*B for each block */
	  cuDoubleComplex ph, pl;
	  int    cindex = RECTB*x+y;
	  CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y] += Axs[x]*Bxp[y];
	}
      }
    }
    __syncthreads();

#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahs[tx2 + y*TILEWIDTH][ty2] = xxAh[y]; 
      Als[tx2 + y*TILEWIDTH][ty2] = xxAl[y]; 
    }
#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+y*TILEWIDTH] = xxBh[y];
      Bls[ty2][tx2+y*TILEWIDTH] = xxBl[y];
    }

    __syncthreads();
  } 

  Ch += tx2 + ibx + __mul24(ty2 + iby ,ldc);
  Cl += tx2 + ibx + __mul24(ty2 + iby ,ldc);

#pragma unroll 
  for(int j1=0; j1<TILEWIDTH; j1++){
#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
      Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhxp[y] = Bhs[j1][ty2 + y*TILEWIDTH];
      Blxp[y] = Bls[j1][ty2 + y*TILEWIDTH];
    }

#pragma unroll 
    for( int x=0; x<RECTA; x++){
#pragma unroll 
      for( int y=0;y<RECTB;y++){
	/* C = A*B for each block */
	cuDoubleComplex ph, pl;
	int    cindex = RECTB*x+y;
	CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	//	Cb[x*RECTA+y]  += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby + ty2;
#pragma unroll
  for( int y=0; y<RECTB; y++, gy+=TILEWIDTH){
    int gx = ibx + tx2; 
#pragma unroll
    for(int x=0; x<RECTA; x++, gx+=TILEWIDTH){
      if (gx < m && gy < n){
	cuDoubleComplex ah, al, bh, bl;
	int    cindex = y+x*RECTB;
	  
	CMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	CMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	CADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] = alpha*Cb[y + x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }

    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}

extern "C" __global__
void gddzgemm_kernel_TN
( int m, int n, int k,
  bool conjA, bool conjB,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 = tx;
  int ty2 = ty;
  int tll = tx2;

  __shared__ cuDoubleComplex Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ cuDoubleComplex Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  cuDoubleComplex xxAh[RECTA], xxAl[RECTA];
  cuDoubleComplex xxBh[RECTB], xxBl[RECTB];

  int trackA = tx2 + __mul24( ibx + ty2*RECTA, lda ); 
  int trackB = tx2 + __mul24( iby + ty2*RECTB, ldb ); 

  if( !conjA ){
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[ty2*RECTA+y][tx2] =
	cuCmul( fetch_x_Ah(trackA + y*lda),
		make_cuDoubleComplex(double(tll<k),0.0) );
      Als[ty2*RECTA+y][tx2] =
	cuCmul( fetch_x_Al(trackA + y*lda),
		make_cuDoubleComplex(double(tll<k),0.0) );
    }
  }
  else{
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[ty2*RECTA+y][tx2] =
	cuCmul( cuConj(fetch_x_Ah(trackA + y*lda)),
		make_cuDoubleComplex(double(tll<k),0.0) );
      Als[ty2*RECTA+y][tx2] =
	cuCmul( cuConj(fetch_x_Al(trackA + y*lda)),
		make_cuDoubleComplex(double(tll<k),0.0) );
    }
  }

  if( !conjB ){
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB+y] = fetch_x_Bh( trackB + y*ldb );
      Bls[tx2][ty2*RECTB+y] = fetch_x_Bl( trackB + y*ldb );
    }
  }
  else{
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB+y] = cuConj(fetch_x_Bh( trackB + y*ldb ));
      Bls[tx2][ty2*RECTB+y] = cuConj(fetch_x_Bl( trackB + y*ldb ));
    }
  }
  __syncthreads();
   
  cuDoubleComplex Ahxs[RECTA], Alxs[RECTA];
  cuDoubleComplex Bhxp[RECTB], Blxp[RECTB];
  cuDoubleComplex Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
      Clb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
    }
  }

  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll +=TILEWIDTH;
    trackA+=TILEWIDTH; 
    trackB+=TILEWIDTH;

    if( !conjA ){
#pragma unroll
      for(int y=0; y<RECTA; y++){
	xxAh[y] =
	  cuCmul( fetch_x_Ah(trackA + y*lda), 
		  make_cuDoubleComplex(double(tll<k),0.0) );
	xxAl[y] =
	  cuCmul( fetch_x_Al(trackA + y*lda), 
		  make_cuDoubleComplex(double(tll<k),0.0) );
      }
    }
    else{
#pragma unroll
      for(int y=0; y<RECTA; y++){
	xxAh[y] =
	  cuCmul( cuConj(fetch_x_Ah(trackA + y*lda)), 
		  make_cuDoubleComplex(double(tll<k),0.0) );
	xxAl[y] =
	  cuCmul( cuConj(fetch_x_Al(trackA + y*lda)), 
		  make_cuDoubleComplex(double(tll<k),0.0) );
      }
    }

    if( !conjB ){
#pragma unroll
      for(int y=0; y<RECTB; y++){
	xxBh[y] = fetch_x_Bh(trackB + y*ldb);
	xxBl[y] = fetch_x_Bl(trackB + y*ldb);
      }
    }
    else{
#pragma unroll
      for(int y=0; y<RECTB; y++){
	xxBh[y] = cuConj(fetch_x_Bh(trackB + y*ldb));
	xxBl[y] = cuConj(fetch_x_Bl(trackB + y*ldb));
      }
    }

#pragma unroll 
    for(int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for(int y=0; y<RECTA; y++){
	Ahxs[y] = Ahs[tx2+y*TILEWIDTH][j1];
	Alxs[y] = Als[tx2+y*TILEWIDTH][j1];
      }
#pragma unroll
      for(int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2+y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2+y*TILEWIDTH];
      }
#pragma unroll 
      for(int x=0; x<RECTA; x++){
#pragma unroll 
	for(int y=0; y<RECTB; y++){
	  /* C = A*B for each block */
	  cuDoubleComplex ph, pl;
	  int    cindex = RECTB*x+y;
	  CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y]  += Axs[x]*Bxp[y];
	}
      }
    }
    __syncthreads();

#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[ty2*RECTA+y][tx2] = xxAh[y];
      Als[ty2*RECTA+y][tx2] = xxAl[y];
    }

#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[tx2][ty2*RECTB+y] =xxBh[y];
      Bls[tx2][ty2*RECTB+y] =xxBl[y];
    }
    __syncthreads();
  }

  Ch += tx2 + ibx  + __mul24 (ty2 + iby ,ldc);
  Cl += tx2 + ibx  + __mul24 (ty2 + iby ,ldc);

#pragma unroll 
  for(int j1=0; j1<TILEWIDTH; j1++){
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahxs[y] = Ahs[tx2+y*TILEWIDTH][j1];
      Alxs[y] = Als[tx2+y*TILEWIDTH][j1];
    }
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhxp[y]= Bhs[j1][ty2+y*TILEWIDTH];
      Blxp[y]= Bls[j1][ty2+y*TILEWIDTH];
    }
#pragma unroll 
    for(int x=0; x<RECTA; x++){
#pragma unroll 
      for(int y=0; y<RECTB; y++){
	/* C = A*B for each block */
	cuDoubleComplex ph, pl;
	int    cindex = RECTB*x+y;
	CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	//	Cb[x*RECTB+y] += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby+ty2;
#pragma unroll
  for(int y=0;y<RECTB;y++, gy+=TILEWIDTH){
    int gx = ibx+tx2;
#pragma unroll
    for(int x=0;x<RECTA;x++, gx+=TILEWIDTH){
      if (gx < m && gy < n){
	cuDoubleComplex ah, al, bh, bl;
	int    cindex = y+x*RECTB;
	  
	CMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	CMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	CADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] =alpha*Cb[y+x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }
    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}


extern "C" __global__
void gddzgemm_kernel_TT
( int m, int n, int k,
  bool conjA, bool conjB,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int ldc )
{
  int tx  = threadIdx.x;
  int ty  = threadIdx.y;
  int ibx = blockIdx.x * BLK_A;
  int iby = blockIdx.y * BLK_B;
  int tx2 = tx;
  int ty2 = ty;
  int tll = tx2;

  __shared__ cuDoubleComplex Ahs[BLK_A][TILEWIDTH+1], Als[BLK_A][TILEWIDTH+1];
  __shared__ cuDoubleComplex Bhs[TILEWIDTH][BLK_B+1], Bls[TILEWIDTH][BLK_B+1];

  cuDoubleComplex xxAh[RECTA], xxAl[RECTA];
  cuDoubleComplex xxBh[RECTB], xxBl[RECTB];

  int trackA = tx2 +__mul24( ibx + ty2, lda);
  int trackB = tx2 + iby + __mul24(ty2, ldb);

  if( !conjA ){
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[ty2+TILEWIDTH*y][tx2] = fetch_x_Ah(trackA +  lda*TILEWIDTH*y);
      Als[ty2+TILEWIDTH*y][tx2] = fetch_x_Al(trackA +  lda*TILEWIDTH*y);
    }
  }
  else{
#pragma unroll
    for(int y=0; y<RECTA; y++){
      Ahs[ty2+TILEWIDTH*y][tx2] = cuConj(fetch_x_Ah(trackA +  lda*TILEWIDTH*y));
      Als[ty2+TILEWIDTH*y][tx2] = cuConj(fetch_x_Al(trackA +  lda*TILEWIDTH*y));
    }
  }

  if( !conjB ){
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+TILEWIDTH*y] = fetch_x_Bh(trackB+TILEWIDTH*y);
      Bls[ty2][tx2+TILEWIDTH*y] = fetch_x_Bl(trackB+TILEWIDTH*y);
    }
  }
  else{
#pragma unroll
    for(int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+TILEWIDTH*y] = cuConj(fetch_x_Bh(trackB+TILEWIDTH*y));
      Bls[ty2][tx2+TILEWIDTH*y] = cuConj(fetch_x_Bl(trackB+TILEWIDTH*y));
    }
  }
  __syncthreads();

  cuDoubleComplex Ahxs[RECTA], Alxs[RECTA];
  cuDoubleComplex Bhxp[RECTB], Blxp[RECTB];
  cuDoubleComplex Chb[RECTA*RECTB], Clb[RECTA*RECTB];

#pragma unroll 
  for( int x=0; x<RECTA; x++){
#pragma unroll 
    for( int y=0; y<RECTB; y++){
      Chb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
      Clb[x*RECTB+y] = make_cuDoubleComplex(0.0,0.0);
    }
  }
        
  int k1;
  for(k1=0; k1<(k-TILEWIDTH); k1+=TILEWIDTH){
    tll+=TILEWIDTH;
    trackA+=TILEWIDTH; 
    trackB+=TILEWIDTH*ldb;
                
    if( !conjA ){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	xxAh[y] = fetch_x_Ah(trackA + lda*y*TILEWIDTH);
	xxAl[y] = fetch_x_Al(trackA + lda*y*TILEWIDTH);
      }
    }
    else{
#pragma unroll
      for( int y=0; y<RECTA; y++){
	xxAh[y] = cuConj(fetch_x_Ah(trackA + lda*y*TILEWIDTH));
	xxAl[y] = cuConj(fetch_x_Al(trackA + lda*y*TILEWIDTH));
      }
    }

    if( !conjB ){
#pragma unroll
      for( int y=0; y<RECTB; y++){
	xxBh[y] = fetch_x_Bh(trackB + TILEWIDTH*y);
	xxBl[y] = fetch_x_Bl(trackB + TILEWIDTH*y);
      }
    }
    else{
#pragma unroll
      for( int y=0; y<RECTB; y++){
	xxBh[y] = cuConj(fetch_x_Bh(trackB + TILEWIDTH*y));
	xxBl[y] = cuConj(fetch_x_Bl(trackB + TILEWIDTH*y));
      }
    }

#pragma unroll 
    for( int j1=0;j1<TILEWIDTH;j1++){
#pragma unroll
      for( int y=0; y<RECTA; y++){
	Ahxs[y] = Ahs[tx2 + y*TILEWIDTH][j1];
	Alxs[y] = Als[tx2 + y*TILEWIDTH][j1];
      }

#pragma unroll
      for( int y=0; y<RECTB; y++){
	Bhxp[y]= Bhs[j1][ty2 + y*TILEWIDTH];
	Blxp[y]= Bls[j1][ty2 + y*TILEWIDTH];
      }

#pragma unroll 
      for( int x=0; x<RECTA; x++){
#pragma unroll 
	for( int y=0;y<RECTB;y++){
	  /* C = A*B for each block */
	  cuDoubleComplex ph, pl;
	  int    cindex = RECTB*x+y;
	  CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	  CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	  //	  Cb[x*RECTB+y] += Axs[x]*Bxp[y];
	}
      }
    }
    __syncthreads();

#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahs[ty2 + TILEWIDTH*y][tx2] = xxAh[y];
      Als[ty2 + TILEWIDTH*y][tx2] = xxAl[y];
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhs[ty2][tx2+y*TILEWIDTH] = xxBh[y];
      Bls[ty2][tx2+y*TILEWIDTH] = xxBl[y];
    }

    __syncthreads();
  } 

  Ch += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);
  Cl += tx2 + ibx  + __mul24 (ty2 +  iby ,ldc);

#pragma unroll 
  for( int j1=0; j1<TILEWIDTH; j1++){
#pragma unroll
    for( int y=0; y<RECTA; y++){
      Ahxs[y] =  Ahs[tx2 + y*TILEWIDTH][j1];
      Alxs[y] =  Als[tx2 + y*TILEWIDTH][j1];
    }

#pragma unroll
    for( int y=0; y<RECTB; y++){
      Bhxp[y] = Bhs[j1][ty2 + y*TILEWIDTH];
      Blxp[y] = Bls[j1][ty2 + y*TILEWIDTH];
    }

#pragma unroll 
    for( int x=0; x<RECTA; x++){
#pragma unroll 
      for( int y=0; y<RECTB; y++){
	/* C = A*B for each block */
	cuDoubleComplex ph, pl;
	int    cindex = RECTB*x+y;
	CMUL( Ahxs+x, Alxs+x, Bhxp+y, Blxp+y, &ph, &pl );
	CADD( Chb+cindex, Clb+cindex, &ph, &pl, Chb+cindex, Clb+cindex );
	//	Cb[x*RECTB+y]  += Axs[x]*Bxp[y];
      }
    }
  }

  int gy = iby + ty2;
#pragma unroll
  for( int y=0; y<RECTB; y++, gy+=TILEWIDTH){
    int gx = ibx + tx2; 
#pragma unroll
    for(int x=0; x<RECTA; x++, gx+=TILEWIDTH){
      if (gx < m && gy < n){
	cuDoubleComplex ah, al, bh, bl;
	int    cindex = y+x*RECTB;
	  
	CMUL( alphah, alphal, Chb+cindex, Clb+cindex, &ah, &al );
	CMUL( betah, betal, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x, &bh, &bl );
	CADD( &ah, &al, &bh, &bl, Ch+TILEWIDTH*x, Cl+TILEWIDTH*x );
	//	C[x*TILEWIDTH] = alpha*Cb[y+x*RECTB] + beta * C[x*TILEWIDTH];
      }
    }

    Ch += ldc*TILEWIDTH;
    Cl += ldc*TILEWIDTH;
  }
}



//----------------------------------------------
extern "C" void gddzgemm
( char* transA, char* transB,
  int* m, int* n, int* k,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int* lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int* ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;

  if( *transA == 'N' || *transA == 'n' ){
    if( *transB == 'N' || *transB == 'n' ){
      // C(M,N) += A(M,K)*B(K,N), A and B are general
      nrowA = *m, ncolA = *k;
      nrowB = *k, ncolB = *n;
      nrowC = *m, ncolC = *n;
    }
    else{
      // C(M,N) += A(M,K)*B(N,K)^T, A and B are general
      nrowA = *m, ncolA = *k;
      nrowB = *n, ncolB = *k;
      nrowC = *m, ncolC = *n;
    }
  }
  else{
    if( *transB == 'N' || *transB == 'n' ){
      // A(K,M)^T *B(K,N), A and B are general
      nrowA = *k, ncolA = *m;
      nrowB = *k, ncolB = *n;
      nrowC = *m, ncolC = *n;
    }
    else{
      // A(K,M)^T *B(K,N)^T, A and B are general
      nrowA = *k, ncolA = *m;
      nrowB = *n, ncolB = *k;
      nrowC = *m, ncolC = *n;
    }
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddzgemm\n");
    printf("wrong lda or ldb or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;

  bool conjA = ( *transA == 'C' || *transA == 'c' );
  bool conjB = ( *transB == 'C' || *transB == 'c' );

  // bind data of matrix A and B to texture memory
  if(
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah,
		      sizeA*sizeof(cuDoubleComplex) ) != cudaSuccess ||
     cudaBindTexture( NULL, texture_Al, (int2*)Al,
		      sizeA*sizeof(cuDoubleComplex) ) != cudaSuccess ||
     cudaBindTexture( NULL, texture_Bh, (int2*)Bh,
		      sizeB*sizeof(cuDoubleComplex) ) != cudaSuccess ||
     cudaBindTexture( NULL, texture_Bl, (int2*)Bl,
		      sizeB*sizeof(cuDoubleComplex) ) != cudaSuccess ){
    printf( "cannot bind to texture \n" );
    return;
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( *m/( BLK_A )+( ( *m )%( BLK_A ) != 0 ),
		 *n/( BLK_B )+( ( *n )%( BLK_B ) != 0 ) );

    if( *transA == 'N' || *transA == 'n' ){
      if( *transB == 'N' || *transB == 'n' ){
	gddzgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *k, conjA, conjB,
	    alphah, alphal, Ah, Al, *lda, Bh, Bl, *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
      else{
	gddzgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *k, conjA, conjB,
	    alphah, alphal, Ah, Al, *lda, Bh, Bl, *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
    }
    else{
      if( *transB == 'N' || *transB == 'n' ){
	gddzgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *k, conjA, conjB,
	    alphah, alphal, Ah, Al, *lda, Bh, Bl, *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
      else{
	gddzgemm_kernel_TT<<< blocks, threads, 0, gddstream>>>
	  ( *m, *n, *k, conjA, conjB,
	    alphah, alphal, Ah, Al, *lda, Bh, Bl, *ldb,
	    betah, betal, Ch, Cl, *ldc );
      }
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );
}



//----------------------------------------------
extern "C" void gddzhemm
( char* side, char* uplo,
  int* m, int* n,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int* lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int* ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  
  if( *side == 'L' || *side == 'l' ){
    // C(M,N) += A(M,M)*B(M,N), A is symmetric, B is general
    nrowA = *m, ncolA = *m;
    nrowB = *m, ncolB = *n;
    nrowC = *m, ncolC = *n;
  }
  else{
    // C(M,N) += B(M,N)*A(N,N), A is symmetric, B is general
    nrowB = *m, ncolB = *n;
    nrowA = *n, ncolA = *n;
    nrowC = *m, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddzhemm\n");
    printf("wrong lda or ldb or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;

  // construct full matrix from triangular/symmetrix matrix
  cuDoubleComplex *Afullh, *Afulll;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**)&Afullh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**)&Afulll ) ){
    cublasFree( Afullh );
    cublasFree( Afulll );
    printf("Device Malloc Error in gddzhemm\n");
    return;
  }
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( nrowA/TILEWIDTH+(nrowA%TILEWIDTH!=0),
		 ncolA/TILEWIDTH+(ncolA%TILEWIDTH!=0) );

    if( *uplo == 'L' || *uplo == 'l' ){
      gddzHELtoFull<<< blocks, threads, 0, gddstream>>>
	( ncolA, Ah, Al, Afullh, Afulll, *lda );
    }
    else{
      gddzHEUtoFull<<< blocks, threads, 0, gddstream>>>
	( ncolA, Ah, Al, Afullh, Afulll, *lda );
    }
  }

  if( *side == 'L' || *side == 'l' ){
    // bind data of matrix A and B to texture memory
    if(
       cudaBindTexture( NULL, texture_Ah, (int2*)Afullh,
			sizeA*sizeof(cuDoubleComplex) ) != cudaSuccess ||
       cudaBindTexture( NULL, texture_Al, (int2*)Afulll,
			sizeA*sizeof(cuDoubleComplex) ) != cudaSuccess ||
       cudaBindTexture( NULL, texture_Bh, (int2*)Bh,
			sizeB*sizeof(cuDoubleComplex) ) != cudaSuccess ||
       cudaBindTexture( NULL, texture_Bl, (int2*)Bl,
			sizeB*sizeof(cuDoubleComplex) ) != cudaSuccess ){
      printf( "cannot bind to texture \n" );
      return;
    }
  }
  else{
    // bind data of matrix B and A to texture memory
    if(
       cudaBindTexture( NULL, texture_Bh, (int2*)Afullh,
			sizeA*sizeof(cuDoubleComplex) ) != cudaSuccess ||
       cudaBindTexture( NULL, texture_Bl, (int2*)Afulll,
			sizeA*sizeof(cuDoubleComplex) ) != cudaSuccess ||
       cudaBindTexture( NULL, texture_Ah, (int2*)Bh,
			sizeB*sizeof(cuDoubleComplex) ) != cudaSuccess ||
       cudaBindTexture( NULL, texture_Al, (int2*)Bl,
			sizeB*sizeof(cuDoubleComplex) ) != cudaSuccess ){
      printf( "cannot bind to texture \n" );
      return;
    }
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*m)/BLK_A+((*m)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    bool conjA = false, conjB = false;

    if( *side == 'L' || *side == 'l' ){
      // C += A*B
      gddzgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	( *m, *n, *m, conjA, conjB,
	  alphah, alphal,
	  Afullh, Afulll, *lda,
	  Bh, Bl, *ldb,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
    else{
      // C += B*A
      gddzgemm_kernel_NN<<< blocks, threads, 0, gddstream>>>
	( *m, *n, *n, conjA, conjB,
	  alphah, alphal,
	  Bh, Bl, *ldb,
	  Afullh, Afulll, *lda,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Afullh );
  cublasFree( Afulll );
}

//----------------------------------------------
extern "C" void gddzsyr2k
( char* uplo, char* trans,
  int* n, int* k,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int* lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int *ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;

  if( *trans == 'N' || *trans == 'n' ){
    // C(N,N) += A(N,K) * B(N,K)^T
    nrowA = *n, ncolA = *k;
    nrowB = *n, ncolB = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    // C(N,N) += A(K,N)^T * B(K,N)
    nrowA = *k, ncolA = *n;
    nrowB = *k, ncolB = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddzsyr2k\n");
    printf("wrong lda or ldb or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;
  int sizeC = (*ldc)*ncolC;

  // bind data of matrix A and B to texture memory
  if(
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Al, (int2*)Al, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bh, (int2*)Bh, sizeB*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bl, (int2*)Bl, sizeB*sizeof(cuDoubleComplex) ) ){
    printf( "cannot bind to texture \n" );
    return;
  }

  // allocate temporal arrays
  cuDoubleComplex *Dh, *Dl;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dl ) ){
    cublasFree( Dh );
    cublasFree( Dl );
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/BLK_A+((*n)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    // make a one on devive
    cuDoubleComplex onehh = make_cuDoubleComplex(1.0,0.0);
    cuDoubleComplex onehl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *oneh, *onel;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&oneh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&onel );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehh, 1, oneh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehl, 1, onel, 1 );

    // make a zero on devive
    cuDoubleComplex zerohh = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex zerohl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *zeroh, *zerol;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zeroh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zerol );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohl, 1, zerol, 1 );

    bool conjA = false, conjB = false;

    if( *trans == 'N' || *trans == 'n' ){
      // D = 1.0*A*B^T + 0.0*D
      gddzgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Bh,  Bl, *ldb,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
    else{
      // D = 1.0*alpha*A^T*B + 0.0*D
      gddzgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Bh,  Bl, *ldb,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }

    cublasFree( zeroh );
    cublasFree( zerol );
    cublasFree( oneh );
    cublasFree( onel );
  }

  // exec extra kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0),
		 (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0) );

    // C = alpha*(D + D^T) + beta*C
    if( *uplo == 'L' || *uplo == 'l' ){
      gddzsyr2k_add_LL <<< blocks, threads, 0, gddstream>>>
	( *n,
	  alphah, alphal,
	  Dh, Dl, *ldc,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
    else{
      gddzsyr2k_add_RR <<< blocks, threads, 0, gddstream>>>
	( *n,
	  alphah, alphal,
	  Dh, Dl, *ldc,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Dh );
  cublasFree( Dl );
}



//----------------------------------------------
extern "C" void gddzher2k
( char* uplo, char* trans,
  int* n, int* k,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int* lda,
  cuDoubleComplex Bh[], cuDoubleComplex Bl[], int *ldb,
  double* betah, double* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;

  if( *trans == 'N' || *trans == 'n' ){
    // C(N,N) += A(N,K) * B(N,K)^T
    nrowA = *n, ncolA = *k;
    nrowB = *n, ncolB = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    // C(N,N) += A(K,N)^T * B(K,N)
    nrowA = *k, ncolA = *n;
    nrowB = *k, ncolB = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddzher2k\n");
    printf("wrong lda or ldb or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeB = (*ldb)*ncolB;
  int sizeC = (*ldc)*ncolC;

  // bind data of matrix A and B to texture memory
  if(
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Al, (int2*)Al, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bh, (int2*)Bh, sizeB*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bl, (int2*)Bl, sizeB*sizeof(cuDoubleComplex) ) ){
    printf( "cannot bind to texture \n" );
    return;
  }

  // allocate temporal arrays
  cuDoubleComplex *Dh, *Dl;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dl ) ){
    cublasFree( Dh );
    cublasFree( Dl );
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/BLK_A+((*n)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    // make a one on devive
    cuDoubleComplex onehh = make_cuDoubleComplex(1.0,0.0);
    cuDoubleComplex onehl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *oneh, *onel;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&oneh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&onel );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehh, 1, oneh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehl, 1, onel, 1 );

    // make a zero on devive
    cuDoubleComplex zerohh = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex zerohl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *zeroh, *zerol;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zeroh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zerol );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohl, 1, zerol, 1 );


    if( *trans == 'N' || *trans == 'n' ){
      // D = 1.0*A*B^T + 0.0*D
    bool conjA = false, conjB = true;
      gddzgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Bh,  Bl, *ldb,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
    else{
      // D = 1.0*alpha*A^T*B + 0.0*D
    bool conjA = true, conjB = false;
      gddzgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah,  Al, *lda,
	  Bh,  Bl, *ldb,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }

    cublasFree( zeroh );
    cublasFree( zerol );
    cublasFree( oneh );
    cublasFree( onel );
  }

  // exec extra kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0),
		 (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0) );

    // C = alpha*(D + D^T) + beta*C
    if( *uplo == 'L' || *uplo == 'l' ){
      gddzher2k_add_LL <<< blocks, threads, 0, gddstream>>>
	( *n,
	  alphah, alphal,
	  Dh, Dl, *ldc,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
    else{
      gddzher2k_add_RR <<< blocks, threads, 0, gddstream>>>
	( *n,
	  alphah, alphal,
	  Dh, Dl, *ldc,
	  betah, betal,
	  Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Dh );
  cublasFree( Dl );
}


//----------------------------------------------
extern "C" void gddzsyrk
( char* uplo, char* trans,
  int* n, int* k,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int* lda,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowC, ncolC;

  if( *trans == 'N' || *trans == 'n' ){
    // C(N,N) += A(N,K) * A(N,K)^T
    nrowA = *n, ncolA = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    // C(N,N) += A(K,N)^T * A(K,N)
    nrowA = *k, ncolA = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldc < nrowC ){
    printf("Argument Error in gddzsyrk_host_\n");
    printf("wrong lda or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeC = (*ldc)*ncolC;

  // bind data of matrix A and B to texture memory
  if(
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Al, (int2*)Al, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bh, (int2*)Ah, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bl, (int2*)Al, sizeA*sizeof(cuDoubleComplex) ) ){
    printf( "cannot bind to texture \n" );
    return;
  }

  // allocate temporal arrays
  cuDoubleComplex *Dh, *Dl;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dl ) ){
    cublasFree( Dh );
    cublasFree( Dl );
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/BLK_A+((*n)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    // make a one on devive
    cuDoubleComplex onehh = make_cuDoubleComplex(1.0,0.0);
    cuDoubleComplex onehl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *oneh, *onel;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&oneh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&onel );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehh, 1, oneh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehl, 1, onel, 1 );

    // make a zero on devive
    cuDoubleComplex zerohh = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex zerohl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *zeroh, *zerol;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zeroh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zerol );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohl, 1, zerol, 1 );

    bool conjA = false, conjB = false;

    if( *trans == 'N' || *trans == 'n' ){
      // D = 1.0*A*A^T + 0.0*D
      gddzgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah, Al, *lda,
	  Ah, Al, *lda,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
    else{
      // D = 1.0*alpha*A^T*A + 0.0*D
      gddzgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah, Al, *lda,
	  Ah, Al, *lda,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
  }

  // exec extra kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0),
		 (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0) );

    // C = alpha*(D + D^T) + beta*C
    if( *uplo == 'L' || *uplo == 'l' ){
      gddzsyrk_add_LL <<< blocks, threads, 0, gddstream>>>
	( *n, alphah, alphal, Dh, Dl, *ldc, betah, betal, Ch, Cl, *ldc );
    }
    else{
      gddzsyrk_add_RR <<< blocks, threads, 0, gddstream>>>
	( *n, alphah, alphal, Dh, Dl, *ldc, betah, betal, Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Dh );
  cublasFree( Dl );
}


//----------------------------------------------
extern "C" void gddzherk
( char* uplo, char* trans,
  int* n, int* k,
  double* alphah, double* alphal,
  cuDoubleComplex Ah[], cuDoubleComplex Al[], int* lda,
  double* betah, double* betal,
  cuDoubleComplex Ch[], cuDoubleComplex Cl[], int* ldc )
{
  // check number of rows, columns and size of matrices
  int nrowA, ncolA;
  int nrowC, ncolC;

  if( *trans == 'N' || *trans == 'n' ){
    // C(N,N) += A(N,K) * A(N,K)^T
    nrowA = *n, ncolA = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    // C(N,N) += A(K,N)^T * A(K,N)
    nrowA = *k, ncolA = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldc < nrowC ){
    printf("Argument Error in gddzherk_host_\n");
    printf("wrong lda or ldc.\n");
    return;
  }

  int sizeA = (*lda)*ncolA;
  int sizeC = (*ldc)*ncolC;

  // bind data of matrix A and B to texture memory
  if(
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Ah, (int2*)Ah, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Al, (int2*)Al, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bh, (int2*)Ah, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess !=
     cudaBindTexture( NULL, texture_Bl, (int2*)Al, sizeA*sizeof(cuDoubleComplex) ) ){
    printf( "cannot bind to texture \n" );
    return;
  }

  // allocate temporal arrays
  cuDoubleComplex *Dh, *Dl;
  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&Dl ) ){
    cublasFree( Dh );
    cublasFree( Dl );
  }

  // exec kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/BLK_A+((*n)%BLK_A!=0), (*n)/BLK_B+((*n)%BLK_B!=0) );

    // make a one on devive
    cuDoubleComplex onehh = make_cuDoubleComplex(1.0,0.0);
    cuDoubleComplex onehl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *oneh, *onel;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&oneh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&onel );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehh, 1, oneh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &onehl, 1, onel, 1 );

    // make a zero on devive
    cuDoubleComplex zerohh = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex zerohl = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex *zeroh, *zerol;
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zeroh );
    cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&zerol );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), &zerohl, 1, zerol, 1 );


    if( *trans == 'N' || *trans == 'n' ){
      // D = 1.0*A*A^H + 0.0*D
    bool conjA = false, conjB = true;
      gddzgemm_kernel_NT<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah, Al, *lda,
	  Ah, Al, *lda,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
    else{
      // D = 1.0*A^H*A + 0.0*D
    bool conjA = true, conjB = false;
      gddzgemm_kernel_TN<<< blocks, threads, 0, gddstream>>>
	( *n, *n, *k, conjA, conjB,
	  oneh, onel,
	  Ah, Al, *lda,
	  Ah, Al, *lda,
	  zeroh, zerol,
	  Dh, Dl, *ldc );
    }
  }

  // exec extra kernel function
  {
    dim3 threads( TILEWIDTH, TILEWIDTH );
    dim3 blocks( (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0),
		 (*n)/TILEWIDTH+((*n)%TILEWIDTH!=0) );

    // C = alpha*(D + D^T) + beta*C
    if( *uplo == 'L' || *uplo == 'l' ){
      gddzherk_add_LL <<< blocks, threads, 0, gddstream>>>
	( *n, alphah, alphal, Dh, Dl, *ldc, betah, betal, Ch, Cl, *ldc );
    }
    else{
      gddzherk_add_RR <<< blocks, threads, 0, gddstream>>>
	( *n, alphah, alphal, Dh, Dl, *ldc, betah, betal, Ch, Cl, *ldc );
    }
  }

  cudaUnbindTexture( texture_Ah );
  cudaUnbindTexture( texture_Al );
  cudaUnbindTexture( texture_Bh );
  cudaUnbindTexture( texture_Bl );

  cublasFree( Dh );
  cublasFree( Dl );
}
