#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddaxpy_host_multigpu_( int *n, double *alphah, double *alphal, double *xh,
                       double *xl, int*incx,
                       double *yh, double *yl,
                       int *incy  )
{
/* NUMBER OF GPUS */
    int   ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int * partition;
    int * pointerOfEachPartition;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );

    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        double *dxh, *dxl, *dyh, *dyl, *dalphah, *dalphal;
        int     tid = omp_get_thread_num( );
        int     xsize = partition[tid]*( *incx );
        int     ysize = partition[tid]*( *incy );
        int     xOffset = pointerOfEachPartition[tid]*( *incx );
        int     yOffset = pointerOfEachPartition[tid]*( *incy );


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ysize, sizeof( double ), (void**) &dyh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ysize, sizeof( double ), (void**) &dyl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dyh );
            cublasFree( dyl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( xsize, sizeof( double ), xh+xOffset, 1, dxh,
                             1 );
            cublasSetVector( xsize, sizeof( double ), xl+xOffset, 1, dxl,
                             1 );
            cublasSetVector( ysize, sizeof( double ), yh+yOffset, 1, dyh,
                             1 );
            cublasSetVector( ysize, sizeof( double ), yl+yOffset, 1, dyl,
                             1 );
            cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 );
            cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 );
        }

        /* Execute gpu vector add */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Cuda Device = %d\n", tid, devid );
#endif
        gddaxpy( &partition[tid],  dalphah, dalphal, dxh, dxl, incx,
                      dyh, dyl,
                      incy );

        /* Get From Device Memory */
        cublasGetVector( ysize, sizeof( double ), dyh, 1, yh+yOffset, 1 );
        cublasGetVector( ysize, sizeof( double ), dyl, 1, yl+yOffset, 1 );
        printf( "%e\n", *( yh+yOffset ) );

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dalphah );
        cublasFree( dalphal );
    }
    free(partition);
    free(pointerOfEachPartition);
    return 0;
}

#ifdef __cplusplus
}
#endif
