#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddgemm_host_multigpu_( char*transA, char*transB, int *m, int*n, int*k,
                       double *alphah, double *alphal, double *Ah,
                       double *Al, int*lda,
                       double *Bh, double *Bl,
                       int *ldb, double *betah, double *betal, double *Ch,
                       double *Cl,
                       int*ldc )
{
    int      btransA = 1, btransB = 1;

    if( *transA == 'N' || *transA == 'n' ) btransA = 0;
    if( *transB == 'N' || *transB == 'n' ) btransB = 0;
    /* NN-> nrow = k,ncol=n */

    /* NUMBER OF GPUS */
    int      ngpu;
    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int *    partitionA, *partitionB;
    int *    sumPartitionA, *sumPartitionB;
    partitionA = (int*) malloc( sizeof( int )*ngpu );
    sumPartitionA = (int*) malloc( sizeof( int )*ngpu );
    partitionB = (int*) malloc( sizeof( int )*ngpu );
    sumPartitionB = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *m, ngpu,  partitionA, sumPartitionA );
    CalcEachArraySize( *n, ngpu,  partitionB, sumPartitionB );

    int      sizeA = ( *m )*( *k );
    int      sizeB = ( *k )*( *n );
    int      sizeC = ( *m )*( *n );

    /* Working Matrix */


    double * hAh, *hAl, *hBh, *hBl, *hCh, *hCl;
    if( !btransA )
    {
        /* Normal */
        if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hAh, sizeA*sizeof( double ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hAl, sizeA*sizeof( double ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }

        copyToSplitMatrix( *m, *k, *lda, ngpu, partitionA, Ah, Al, hAh, hAl );
    }
    else
    {
        /* Transpose*/
        hAh = Ah;
        hAl = Al;
    }


    if( btransB )
    {
        /* Transpose */
        if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hBh, sizeB*sizeof( double ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hBl, sizeB*sizeof( double ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }

        copyTranspose( *k, *n, *ldb, Bh, Bl, hBh, hBl );
    }
    else
    {
        /* Normal */
        hBh = Bh;
        hBl = Bl;
    }


    if(
        cudaSuccess != cudaMallocHost( (void **)&hCh, sizeC*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&hCl, sizeC*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }
    copyToSplitMatrix( *m, *n, *ldc, ngpu, partitionA, Ch, Cl, hCh, hCl );


    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        int     tid = omp_get_thread_num( );

        int     localldb;

	int localn = partitionB[tid];
        if( btransB )
        {
            localldb = *k;
        }
        else
        {
            localldb = *ldb;
        }

        int     lsizeB = localldb*localn;
        int     offsetB = localldb*sumPartitionB[tid];

        double *dBh, *dBl, *dalphah, *dalphal, *dbetah, *dbetal;

	if(lsizeB != 0)
	{

        /* B,alpha及びbetaを確保 */
        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeB, sizeof( double ), (void**) &dBh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeB, sizeof( double ), (void**) &dBl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dbetah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dbetal )
          )
        {
            /*	 Memory allocation failed */
            cublasFree( dBh );
            cublasFree( dBl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            cublasFree( dbetah );
            cublasFree( dbetal );
            exit( -1 );
        }


        cublasSetVector( lsizeB, sizeof( double ), hBh+offsetB, 1, dBh, 1 );
        cublasSetVector( lsizeB, sizeof( double ), hBl+offsetB, 1, dBl, 1 );
        cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 );
        cublasSetVector( 1, sizeof( double ), betah, 1, dbetah, 1 );
        cublasSetVector( 1, sizeof( double ), betal, 1, dbetal, 1 );


        int idx;
        for( idx = 0; idx < ngpu; idx++ )
        {
            int locallda;
	    int localm = partitionA[idx];
            int localk;
            if( !btransA )
            {
                locallda = localm;
                localk = ( *k );
            }
            else
            {
                locallda = *lda;
                localk = locallda;
            }
            int     offsetA = localk*sumPartitionA[idx];
            int     lsizeA    = localk*localm;

            int     localldc = localm;
            int     lsizeC = localm*localn;
            int     offsetC = sumPartitionB[tid]*localm+
                               sumPartitionA[idx]*( *n );

/*            printf(*/
/*                "idx:%d,tid:%d,locallda:%d,lsizeA:%d,offsetA:%d,localldc:%d,lsizeC:%d,offsetC:%d,pA:%d\n",*/
/*                idx, tid, locallda, lsizeA, offsetA, localldc, lsizeC,*/
/*                offsetC,*/
/*                localm );*/

	    if(lsizeA!=0 && lsizeC != 0)
	    {

            double *dAh, *dAl, *dCh, *dCl;


            if(
                CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeA, sizeof( double ), (void**) &dAh )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeA, sizeof( double ), (void**) &dAl )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( double ), (void**) &dCh )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( double ), (void**) &dCl )
              )
            {
                /*	 Memory allocation failed */
                cublasFree( dAh );
                cublasFree( dAl );
                cublasFree( dCh );
                cublasFree( dCl );
                exit( -1 );
            }


            cublasSetVector( lsizeA, sizeof( double ), hAh+offsetA, 1, dAh,
                             1 );
            cublasSetVector( lsizeA, sizeof( double ), hAl+offsetA, 1, dAl,
                             1 );
            cublasSetVector( lsizeC, sizeof( double ), hCh+offsetC, 1, dCh,
                             1 );
            cublasSetVector( lsizeC, sizeof( double ), hCl+offsetC, 1, dCl,
                             1 );

            /* Execute gpu gemm*/
            char localtransA, localtransB;
            if( !btransA )
            {
                localtransA = 'N';
                localtransB = 'N';
            }
            else
            {
                localtransA = 'T';
                localtransB = 'N';
            }

#ifdef _DEBUG_
            int devid;
            cudaGetDevice( &devid );
            printf( "TILE INDEX(%d,%d), Cuda Device = %d\n", idx, tid,
                    devid );
#endif


            gddgemm( &localtransA, &localtransB,  &localm,  &localn,  &localk, dalphah,
                          dalphal,
                          dAh,  dAl,  &locallda, dBh,  dBl, &localldb,
                          dbetah, dbetal,
                          dCh,  dCl,
                          &localldc );
            cublasGetVector( lsizeC, sizeof( double ), dCh, 1, hCh+offsetC,
                             1 );
            cublasGetVector( lsizeC, sizeof( double ), dCl, 1, hCl+offsetC,
                             1 );

            cublasFree( dAh );
            cublasFree( dAl );
            cublasFree( dCh );
            cublasFree( dCl );
	    }
        }


        cublasFree( dBh );
        cublasFree( dBl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dbetah );
        cublasFree( dbetal );
	}
    }
/*    for (int i = 0; i < *m*(*n); i++) {*/
/*            printf("%e\n",*(hCh+i));*/
/*    }*/

    copyFromSplitMatrix( *m, *n, *ldc, ngpu, partitionA, hCh, hCl, Ch, Cl );

    if(!btransA)
    {
    cudaFreeHost( hAh );
    cudaFreeHost( hAl );
    }
    if(btransB)
    {
    cudaFreeHost( hBh );
    cudaFreeHost( hBl );
    }
    cudaFreeHost( hCh );
    cudaFreeHost( hCl );

    free(partitionA);
    free(partitionB);
    free(sumPartitionA);
    free(sumPartitionB);

    return 0;
}

#ifdef __cplusplus
}
#endif
