#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddrotm_host_multigpu_( int *n, double *xh,
                       double *xl, int*incx,
                       double *yh, double *yl,
                       int *incy,
                       double*paramh, double*paraml )
{
/* NUMBER OF GPUS */
    int   ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int * partition;
    int * pointerOfEachPartition;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );

    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        double *dxh, *dxl, *dyh, *dyl, *dparamh, *dparaml ;
        int     tid = omp_get_thread_num( );
        int     xsize = partition[tid]*( *incx );
        int     ysize = partition[tid]*( *incy );
        int     xOffset = pointerOfEachPartition[tid]*( *incx );
        int     yOffset = pointerOfEachPartition[tid]*( *incy );


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ysize, sizeof( double ), (void**) &dyh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ysize, sizeof( double ), (void**) &dyl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 5, sizeof( double ), (void**) &dparamh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 5, sizeof( double ), (void**) &dparaml )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dyh );
            cublasFree( dyl );
            cublasFree( dparamh );
            cublasFree( dparaml );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( xsize, sizeof( double ), xh+xOffset, 1, dxh,
                             1 );
            cublasSetVector( xsize, sizeof( double ), xl+xOffset, 1, dxl,
                             1 );
            cublasSetVector( ysize, sizeof( double ), yh+yOffset, 1, dyh,
                             1 );
            cublasSetVector( ysize, sizeof( double ), yl+yOffset, 1, dyl,
                             1 );
            cublasSetVector( 5, sizeof( double ), paramh, 1, dparamh, 1 );
            cublasSetVector( 5, sizeof( double ), paraml, 1, dparaml, 1 );
        }

        /* Execute gpu vector add */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Cuda Device = %d\n", tid, devid );
#endif
        gddrotm( &partition[tid],   dxh, dxl, incx,
                     dyh, dyl, incy, dparamh, dparaml );

        /* Get From Device Memory */
        cublasGetVector( xsize, sizeof( double ), dxh, 1, xh+xOffset, 1 );
        cublasGetVector( xsize, sizeof( double ), dxl, 1, xl+xOffset, 1 );
        cublasGetVector( ysize, sizeof( double ), dyh, 1, yh+yOffset, 1 );
        cublasGetVector( ysize, sizeof( double ), dyl, 1, yl+yOffset, 1 );

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dparamh );
        cublasFree( dparaml );
    }
    free( partition );
    free( pointerOfEachPartition );
    return 0;
}

#ifdef __cplusplus
}
#endif
