#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddsyr2k_host_multigpu_( char* uplo, char* trans,
                        int* n, int* k,
                        double* alphah, double* alphal,
                        double* Ah, double* Al, int* lda,
                        double* Bh, double* Bl, int* ldb,
                        double* betah, double* betal,
                        double* Ch, double* Cl, int* ldc )
{
    int      btrans = 0;

    if( *trans == 'T' || *trans == 't' ) btrans = 1;

    /* NUMBER OF GPUS */
    int      ngpu;
    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int *    partition;
    int *    sumPartition;
    partition = (int*) malloc( sizeof( int )*ngpu );
    sumPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, sumPartition );

    int      sizeM = ( *n )*( *k );
    int      sizeC = ( *n )*( *n );

    /* Working Matrix */


    double * hAh, *hAl, *hBh, *hBl, *hCh, *hCl;
    if( !btrans )
    {
        if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hAh, sizeM*sizeof( double ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hAl, sizeM*sizeof( double ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hBh, sizeM*sizeof( double ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hBl, sizeM*sizeof( double ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }

        copyToSplitMatrix( *n, *k, *lda, ngpu, partition, Ah, Al, hAh, hAl );
        copyTranspose( *k, *n, *ldb, Bh, Bl, hBh, hBl );
    }
    else
    {
        hAh = Ah;
        hAl = Al;
        hBh = Bh;
        hBl = Bl;
    }


    if(
        cudaSuccess != cudaMallocHost( (void **)&hCh, sizeC*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&hCl, sizeC*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }
    copyToSymmetricSplitMatrix( *n, *ldc, uplo, ngpu, partition, Ch, Cl,
                                hCh,
                                hCl );


    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        int     tid = omp_get_thread_num( );

        int     localldb = btrans ? ( *ldb ) : ( *k );
        int     localn = partition[tid];

        double *dAh[ngpu], *dAl[ngpu], *dBh[ngpu], *dBl[ngpu], *dalphah, *dalphal,
        *dbetah, *dbetal,
        *doneh, *donel;

        /* B,alpha及びbetaを確保 */
        for( int i = 0; i < ngpu; i++ )
        {
            int sizeA = btrans ? ( *lda )*partition[i] : ( *k )*
                        partition[i];
            int sizeB = btrans ? ( *ldb )*partition[i] : ( *k )*
                        partition[i];

            if(
                CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeA, sizeof( double ), (void**) &dAh[i] )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeA, sizeof( double ), (void**) &dAl[i] )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeB, sizeof( double ), (void**) &dBh[i] )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeB, sizeof( double ), (void**) &dBl[i] )
              )
            {
                exit( -1 );
            }
        }
        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dbetah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dbetal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &doneh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &donel )
          )
        {
            cublasFree( dalphah );
            cublasFree( dalphal );
            cublasFree( dbetah );
            cublasFree( dbetal );
            cublasFree( doneh );
            cublasFree( donel );
            exit( -1 );
        }

        double oneh = 1.0;
        double onel = 0.0;

        for( int i = 0; i < ngpu; i++ )
        {
            int offsetA = btrans ? ( *lda )*sumPartition[i] : ( *k )*
                          sumPartition[i];
            int offsetB = btrans ? ( *ldb )*sumPartition[i] : ( *k )*
                          sumPartition[i];
            int sizeA = btrans ? ( *lda )*partition[i] : ( *k )*
                        partition[i];
            int sizeB = btrans ? ( *ldb )*partition[i] : ( *k )*
                        partition[i];
            cublasSetVector( sizeA, sizeof( double ), hAh+offsetA, 1,
                             dAh[i],
                             1 );
            cublasSetVector( sizeA, sizeof( double ), hAl+offsetA, 1,
                             dAl[i],
                             1 );
            cublasSetVector( sizeB, sizeof( double ), hBh+offsetB, 1,
                             dBh[i],
                             1 );
            cublasSetVector( sizeB, sizeof( double ), hBl+offsetB, 1,
                             dBl[i],
                             1 );
        }
        cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 );
        cublasSetVector( 1, sizeof( double ), betah, 1, dbetah, 1 );
        cublasSetVector( 1, sizeof( double ), betal, 1, dbetal, 1 );
        cublasSetVector( 1, sizeof( double ), &oneh, 1, doneh, 1 );
        cublasSetVector( 1, sizeof( double ), &onel, 1, donel, 1 );

        int idx;
        for( idx = 0; idx < ngpu; idx++ )
        {
            int     localm = partition[idx];
            int     localk = *k;


            int     localldc = localm;
            int     lsizeC = localm*localn;
            int     offsetC = sumPartition[tid]*localm+
                              sumPartition[idx]*( *n );

            double *dCh, *dCl;

            if(
                CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( double ), (void**) &dCh )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( double ), (void**) &dCl )
              )
            {
                /*	 Memory allocation failed */
                cublasFree( dCh );
                cublasFree( dCl );
                exit( -1 );
            }


            cublasSetVector( lsizeC, sizeof( double ), hCh+offsetC, 1, dCh,
                             1 );
            cublasSetVector( lsizeC, sizeof( double ), hCl+offsetC, 1, dCl,
                             1 );

#ifdef _DEBUG_
            int  devid;
            cudaGetDevice( &devid );
            printf( "TILE INDEX(%d,%d), Cuda Device = %d\n", idx, tid,
                    devid );
#endif


            char localtransA = btrans ? 'T' : 'N';
            char localtransB = btrans ? 'N' : 'N';

            int  locallda = btrans ? ( *lda ) : partition[idx];

            /* Execute gpu gemm*/

            gddgemm( &localtransA, &localtransB,  &localm,  &localn,
                          &localk, dalphah,
                          dalphal,
                          dAh[idx],  dAl[idx],  &locallda, dBh[tid],
                          dBl[tid], &localldb,
                          dbetah, dbetal,
                          dCh,  dCl,
                          &localldc );

            localtransA = btrans ? 'N' : 'T';
            localtransB = btrans ? 'T' : 'T';
            locallda = btrans ? ( *lda ) : partition[tid];

            gddgemm( &localtransB, &localtransA,  &localm,  &localn,
                          &localk, dalphah,
                          dalphal,
                          dBh[idx],  dBl[idx],  &localldb, dAh[tid],
                          dAl[tid], &locallda,
                          doneh, donel,
                          dCh,  dCl,
                          &localldc );


            cublasGetVector( lsizeC, sizeof( double ), dCh, 1, hCh+offsetC,
                             1 );
            cublasGetVector( lsizeC, sizeof( double ), dCl, 1, hCl+offsetC,
                             1 );

            cublasFree( dCh );
            cublasFree( dCl );
        }

        for( int i = 0; i < ngpu; i++ )
        {
            cublasFree( dAh[i] );
            cublasFree( dAl[i] );
        }

        cublasFree( dBh );
        cublasFree( dBl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dbetah );
        cublasFree( dbetal );
        cublasFree( doneh );
        cublasFree( donel );
    }


    copyFromSymmetricSplitMatrix( *n, *ldc, uplo, ngpu, partition, hCh, hCl,
                                  Ch,
                                  Cl );


    if( !btrans )
    {
        cudaFreeHost( hAh );
        cudaFreeHost( hAl );
        cudaFreeHost( hBh );
        cudaFreeHost( hBl );
    }
    cudaFreeHost( hCh );
    cudaFreeHost( hCl );

    free( partition );
    free( sumPartition );

    return 0;
}

#ifdef __cplusplus
}
#endif
