#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddsyr_host_multigpu_( char* uplo, int *n, double *alphah, double*alphal,
                      double * xh, double *xl,
                      int *incx,
                      double *Ah, double *Al,
                      int *lda )
{
/* NUMBER OF GPUS */
    int      ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int *    partition;
    int *    pointerOfEachPartition;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );


    double *hAh,*hAl;
    int matsize = (*n)* (*n);
    if(
        cudaSuccess != cudaMallocHost(  (void **)&hAh, matsize*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&hAl, matsize*sizeof( double ) )
      )
    {
    cudaFreeHost( hAh );
    cudaFreeHost( hAl );
        printf( "Host Malloc Error\n" );
        return -1;
    }




    copyToSymmetricMatrix(*n,*lda,uplo,Ah,Al,hAh,hAl);

    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        double *dxh, *dxl, *dAh, *dAl,*dalphah,*dalphal;
        int     tid = omp_get_thread_num( );
        int     sizex = ( *n )*( *incx );
	int     offsetx = pointerOfEachPartition[tid]*( *incx );
	int     sizeA = (*n)*partition[tid];
	int     offsetA = pointerOfEachPartition[tid]*(*n);


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( double ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( double ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizeA, sizeof( double ), (void**) &dAh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizeA, sizeof( double ), (void**) &dAl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dAh );
            cublasFree( dAl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( sizex, sizeof( double ), xh, 1, dxh,1 );
            cublasSetVector( sizex, sizeof( double ), xl, 1, dxl,1 );
            cublasSetVector( sizeA, sizeof( double ), hAh+offsetA, 1, dAh,1 );
            cublasSetVector( sizeA, sizeof( double ), hAl+offsetA, 1, dAl,1 );
            cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah,1 );
            cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal,1 );
        }


        /* Execute gpu vector add */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Cuda Device = %d\n", tid, devid );
#endif
        gddger( n,&partition[tid], dalphah,dalphal, dxh, dxl, incx,
                     dxh+offsetx, dxl+offsetx, incx,
                     dAh,dAl,lda );

        /* Get From Device Memory */
        cublasGetVector( sizeA, sizeof( double ), dAh, 1, hAh+offsetA, 1 );
        cublasGetVector( sizeA, sizeof( double ), dAl, 1, hAl+offsetA, 1 );

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dAh );
        cublasFree( dAl );
        cublasFree( dalphah );
        cublasFree( dalphal );
    }



    
    copyFromSymmetricMatrix(*n,*lda,uplo,hAh,hAl,Ah,Al);

    free( partition );
    free( pointerOfEachPartition );
    cudaFreeHost( hAh );
    cudaFreeHost( hAl );
    return 0;
}

#ifdef __cplusplus
}
#endif
