#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddtrmv_host_multigpu_( char*uplo, char*trans, char*diag, int *n,
                        double *Ah, double *Al, int *lda, double * xh,
                        double *xl,
                        int *incx   )
{
    /* NUMBER OF GPUS */
    int     ngpu;

    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    int     btrans = 0;
    if( *trans == 'T' || *trans == 't' ) btrans = 1;

    int     bupflag = 0;
    if( *uplo == 'U' || *uplo == 'u' ) bupflag = 1;


    int     sizex = *n;


    /* ARRAYSIZE FOR EACH GPU */
    int *   partition;
    int *   sumPartition;
    partition =    (int*) malloc( sizeof( int )*ngpu );
    sumPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( sizex, ngpu,  partition, sumPartition );


    int     sizeA = ( *n )*( *n );

    double *hAh, *hAl;
    if(
        cudaSuccess !=
        cudaMallocHost( (void **)&hAh, sizeA*sizeof( double ) )
      || cudaSuccess !=
        cudaMallocHost( (void **)&hAl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    char localuplo;
    if( btrans )
    {
        if( bupflag )
        {
            localuplo = 'l';
        }
        else
        {
            localuplo  = 'u';
        }
    }
    else
    {
        localuplo = *uplo;
    }
    copyToTriangleSplitMatrix( &localuplo, trans, diag, *n, *lda, ngpu,
                               partition, Ah, Al, hAh,
                               hAl );

/*    int i, j, k;*/

/*    for( i = 0; i < ngpu; i++ )*/
/*    {*/
/*        for( j = 0; j < partition[i]; j++ )*/
/*        {*/
/*            for( k = 0; k < *n; k++ )*/
/*            {*/
/*                printf( "%e,", *( hAh+j+k*partition[i]+sumPartition[i]*( *n ) ) );*/
/*            }*/
/*            printf( "\n" );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

    double alphah = 1.0;
    double alphal = 0.0;
    double betah = 0.0;
    double betal = 0.0;


/*     Working Matrix  */
    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        int     tid = omp_get_thread_num( );

        double *dAh, *dAl, *dxh, *dxl, *dworkh, *dworkl;

        double *dalphah, *dalphal, *dbetah, *dbetal;


        int     offsetx = sumPartition[tid]*( *incx );
        int     localsizex = partition[tid];

        int     offsetA = sumPartition[tid]*( *n );
        int     localsizeA = partition[tid]*( *n );
        int     locallda = partition[tid];

        int     localm = partition[tid];
        int     localn = *n;

        /* B,alpha及びbetaを確保 */
        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( double ), (void**) &dworkh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( double ), (void**) &dworkl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizex, sizeof( double ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizex, sizeof( double ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dbetah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dbetal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizeA, sizeof( double ), (void**) &dAh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizeA, sizeof( double ), (void**) &dAl )
          )
        {
            /*	 Memory allocation failed */
            cublasFree( dworkh );
            cublasFree( dworkl );
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            cublasFree( dbetah );
            cublasFree( dbetal );
            cublasFree( dAh );
            cublasFree( dAl );
            exit( -1 );
        }


        cublasSetVector( sizex, sizeof( double ), xh, *incx, dworkh, 1 );
        cublasSetVector( sizex, sizeof( double ), xl, *incx, dworkl, 1 );
        cublasSetVector( localsizeA, sizeof( double ), hAh+offsetA, 1, dAh,
                         1 );
        cublasSetVector( localsizeA, sizeof( double ), hAl+offsetA, 1, dAl,
                         1 );
        cublasSetVector( 1, sizeof( double ), &alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( double ), &alphal, 1, dalphal, 1 );
        cublasSetVector( 1, sizeof( double ), &betah, 1, dbetah, 1 );
        cublasSetVector( 1, sizeof( double ), &betal, 1, dbetal, 1 );


#ifdef _DEBUG_
        int  devid;
        cudaGetDevice( &devid );
        printf( "TILE INDEX(%d), Cuda Device = %d\n", tid,
                devid );
#endif
        char localtrans = 'N';


        int  localinc = 1;
        gddgemv( &localtrans,  &localm,  &localn, dalphah,  dalphal, dAh,
                 dAl,
                 &locallda, dworkh, dworkl,  &localinc,  dbetah, dbetal,
                 dxh,  dxl,
                 &localinc );
#pragma omp barrier
        cublasGetVector( localsizex, sizeof( double ), dxh, 1, xh+offsetx,
                         *incx );
        cublasGetVector( localsizex, sizeof( double ), dxl, 1, xl+offsetx,
                         *incx );

        cublasFree( dAh );
        cublasFree( dAl );
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dworkh );
        cublasFree( dworkl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dbetah );
        cublasFree( dbetal );
    }


    cudaFreeHost( hAh );
    cudaFreeHost( hAl );

    free( partition );
    free( sumPartition );

    return 0;
}

#ifdef __cplusplus
}
#endif
