#include "gddcommon.h"
#include "multiple_gpu.h"
#include "gddblas_host.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#define BLOCKSIZE ( 128 )

#ifdef __cplusplus
extern "C" {
#endif


static __global__ void
setUnitMatrix_kernel_( int n, double *Ah, double*Al, int lda )
{
    int idx = blockIdx.x*blockDim.x+threadIdx.x;

    for( int i = 0; i < n; i++ )
    {
        int flag = ( idx == i );
        if( idx < n )
        {
            *( Ah+idx+lda*i ) = (double) flag;
            *( Al+idx+lda*i ) = 0.0;
        }
    }
}

void
gddSetUnitMatrix(  int* n, double Ah[], double Al[], int* lda )
{
    dim3 threads( BLOCKSIZE );

    dim3 blocks( ( *n )/BLOCKSIZE+( ( *n )%BLOCKSIZE != 0 ) );

    setUnitMatrix_kernel_<<< blocks, threads, 0, gddstream>>>
    ( *n, Ah, Al, *lda );
}

int
gddtrsm_host_multigpu_( char* side, char* uplo, char* transA, char* diag,
                       int* m, int* n,
                       double* alphah, double* alphal,
                       double* Ah, double* Al, int* lda,
                       double* Bh, double* Bl, int* ldb )
{
    /* NUMBER OF GPUS */
    int ngpu;

    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    int nrowA, ncolA;
    int nrowB, ncolB;
    int sizeA, sizeB;

    nrowB = *m, ncolB = *n;
    int bleftside = 0;
    if( *side == 'L' || *side == 'l' ) bleftside = 1;
    if( bleftside )
    {
        nrowA = *m, ncolA = *m;
    }
    else
    {
        nrowA = *n, ncolA = *n;
    }


    /* ARRAYSIZE FOR EACH GPU */
    int * partition;
    int * sumPartition;
    partition = (int*) malloc( sizeof( int )*ngpu );
    sumPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( nrowA, ngpu,  partition, sumPartition );


    int   bupside = 0;
    if( *uplo == 'U' || *uplo == 'u' ) bupside = 1;

    int   btransA = 0;
    if( *transA == 'T' || *transA == 't' ) btransA = 1;

    sizeA = nrowA*ncolA;

    if( bleftside )
    {
        sizeB = nrowB*ncolB;
    }
    else
    {
        sizeB = ( *ldb )*ncolB;
    }

    /* Working Matrix */

    double * hAh, *hAl, *hBh, *hBl;

    if(
        cudaSuccess !=
        cudaMallocHost( (void **)&hAh, sizeA*sizeof( double ) )
      || cudaSuccess !=
        cudaMallocHost( (void **)&hAl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    char tmptrans = 'N';
    copyToTriangleSplitMatrix( uplo, &tmptrans, diag, nrowA, *lda, ngpu,
                               partition, Ah, Al, hAh,
                               hAl );
#ifdef _DEBUG_
/*    int i, j, k;*/

/*    for( i = 0; i < ngpu; i++ )*/
/*    {*/
/*        for( j = 0; j < partition[i]; j++ )*/
/*        {*/
/*            for( k = 0; k < ncolA; k++ )*/
/*            {*/
/*                printf( "%e,",*/
/*                       *( hAh+j+k*partition[i]+sumPartition[i]*( ncolA ) ));*/
/*            }*/
/*            printf( "\n" );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/
#endif


    if( bleftside )
    {
        if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hBh, sizeB*sizeof( double ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hBl, sizeB*sizeof( double ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }

        copyToSplitMatrix( nrowB, ncolB, *ldb, ngpu, partition, Bh, Bl, hBh,
                           hBl );
    }
    else
    {
        hBh = Bh;
        hBl = Bl;
    }

    static  double oneh = 1.0, onel = 0.0;
    static  double moneh = -1.0, monel = 0.0;
    static  double zeroh = 0.0, zerol = 0.0;


    /* calculate the inverse of diagonal blocks */
    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        int     tid = omp_get_thread_num( );
        int     offsetA = sumPartition[tid]*ncolA+sumPartition[tid]*
                          partition[tid];
        int     ln = partition[tid];
        int     lsizeA  = ln*ln;

        double *dAh, *dAl, *dINVh, *dINVl, *doneh, *donel;

        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeA, sizeof( double ), (void**) &dAh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeA, sizeof( double ), (void**) &dAl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeA, sizeof( double ), (void**) &dINVh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeA, sizeof( double ), (void**) &dINVl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &doneh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &donel )
          )
        {
            /*	 Memory allocation failed */
            cublasFree( dAh );
            cublasFree( dAl );
        cublasFree( doneh );
        cublasFree( donel );
            cublasFree( dINVh );
            cublasFree( dINVl );
            exit( -1 );
        }

        cublasSetVector( lsizeA, sizeof( double ), hAh+offsetA, 1, dAh, 1 );
        cublasSetVector( lsizeA, sizeof( double ), hAl+offsetA, 1, dAl, 1 );
        cublasSetVector( 1, sizeof( double ), &oneh, 1, doneh, 1 );
        cublasSetVector( 1, sizeof( double ), &onel, 1, donel, 1 );
        gddSetUnitMatrix( &ln, dINVh, dINVl, &ln );

#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TILE INDEX(%d), Cuda Device = %d\n",  tid,
                devid );
#endif


        gddtrsm( side, uplo, &tmptrans, diag,
                      &ln, &ln,
                      doneh, donel,
                      dAh, dAl, &ln,
                      dINVh, dINVl, &ln );

        cublasGetVector( lsizeA, sizeof( double ), dINVh, 1, hAh+offsetA, 1 );
        cublasGetVector( lsizeA, sizeof( double ), dINVl, 1, hAl+offsetA, 1 );

        cublasFree( dAh );
        cublasFree( dAl );
        cublasFree( doneh );
        cublasFree( donel );
        cublasFree( dINVh );
        cublasFree( dINVl );
    }

#ifdef _DEBUG_
/*    for( int i = 0; i < ngpu; i++ )*/
/*    {*/
/*        int ln = partition[i];*/
/*        int offsetA = sumPartition[i]*ncolA+sumPartition[i]*ln;*/
/*        for( int x = 0; x < ln; x++ )*/
/*        {*/
/*            for( int y = 0; y < ln; y++ )*/
/*            {*/
/*                printf( "%e,", *( hAh+offsetA+x+ln*y ) );*/
/*            }*/
/*            printf( "\n" );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/
#endif
    

    for( int i = 0; i < ngpu; i++ )
    {
        int offsetB;
        if( ( !bupside && !btransA && bleftside )
          || ( bupside && btransA && bleftside )
            )
        {
            offsetB =  sumPartition[i]*ncolB;
        }
        else if( ( !bupside && btransA && bleftside )
               || ( bupside && !btransA && bleftside ) )
        {
            offsetB = sumPartition[ngpu-i-1]*ncolB;
        }
        else if( ( bupside && !btransA )
               || ( !bupside && btransA ) )
        {
            offsetB = sumPartition[i]*( *ldb );
        }
        else if( ( !bupside && !btransA )
               || ( bupside && btransA ) )
        {
            offsetB = sumPartition[ngpu-i-1]*( *ldb );
        }


        for( int j = 0; j <= i; j++ )
        {
            if( bleftside )
            {
                int offsetA;
                int idx, jdx;
                if( ( bupside && !btransA ) || ( !bupside && btransA ) )
                {
                    idx = ngpu-i-1;
                    jdx = ngpu-j-1;
                }
                else
                {
                    idx = i;
                    jdx = j;
                }


                int tilewidth = partition[jdx];
                int tileheight = partition[idx];
                int rowidx = sumPartition[idx];
                int colidx = sumPartition[jdx];


                int offsetx;
                if( ( !bupside && !btransA  )
                  || ( bupside && btransA   )
                    )
                {
                    offsetx =  sumPartition[j]*ncolB;
                }
                else if( ( !bupside && btransA   )
                       || ( bupside && !btransA  ) )
                {
                    offsetx = sumPartition[ngpu-j-1]*ncolB;
                }


                if( !bupside && !btransA  )
                {
                    offsetA = rowidx*ncolA
                              +colidx*tileheight;
                }
                else if( bupside && !btransA )
                {
                    offsetA = rowidx*ncolA
                              +colidx*tileheight;
                }
                else if( !bupside && btransA )
                {
                    offsetA = colidx*ncolA
                              +rowidx*tilewidth;
                }
                else if( bupside && btransA )
                {
                    offsetA = colidx*ncolA
                              +rowidx*tilewidth;
                }

                char transB = 'N';


/*

   printf("i:%d,j:%d,offsetA:%d,offsetB:%d,offsetx:%d,tileheight:%d,tilewidth:%d\n",*/
/*

                            i,j,offsetA,offsetB,offsetx,tileheight,tilewidth);*/

                if( i == j )
                {
                    /* B = A_{ii}^(-1)B */

                    gddgemm_host_multigpu_( transA, &transB, &tileheight,
                                           &ncolB, &tilewidth,
                                           alphah, alphal,
                                           hAh+offsetA, hAl+offsetA,
                                           &tileheight,
                                           hBh+offsetB, hBl+offsetB,
                                           &tilewidth,
                                           &zeroh, &zerol,
                                           hBh+offsetB, hBl+offsetB,
                                           &tilewidth );
                }
                else
                {
/*  printf("hAh:%e,hBhb:%e,bBhx:%e\n",*(hAh+offsetA+1),*/
/*                  *(hBh+offsetB+1),*(hBh+offsetx+1));*/
                    if( !btransA )
                    {
                        gddgemm_host_multigpu_( transA, &transB, &tileheight,
                                               &ncolB, &tilewidth,
                                               &moneh, &monel,
                                               hAh+offsetA, hAl+offsetA,
                                               &tileheight,
                                               hBh+offsetx, hBl+offsetx,
                                               &tilewidth,
                                               &oneh, &onel,
                                               hBh+offsetB, hBl+offsetB,
                                               &tileheight );
                    }
                    else
                    {
                        gddgemm_host_multigpu_( transA, &transB, &tileheight,
                                               &ncolB, &tilewidth,
                                               &moneh, &monel,
                                               hAh+offsetA, hAl+offsetA,
                                               &tilewidth,
                                               hBh+offsetx, hBl+offsetx,
                                               &tilewidth,
                                               &oneh, &onel,
                                               hBh+offsetB, hBl+offsetB,
                                               &tileheight );
                    }
                }
            }
            else
            {
                /* rightside */

                int offsetA;
                int idx, jdx;

                if( ( !bupside && !btransA )
                  || ( bupside && btransA ) )
                {
                    idx = ngpu-i-1;
                    jdx = ngpu-j-1;
                }
                else
                {
                    idx = i;
                    jdx = j;
                }


		int tilewidth  = partition[idx];
		int tileheight = partition[jdx];
		int rowidx     = sumPartition[jdx];
		int colidx     = sumPartition[idx];


		int offsetx;
                if( ( !bupside && !btransA )
                  || ( bupside && btransA ) )
		{
			offsetx = sumPartition[ngpu-j-1]*(*ldb);
		}
		else
		{
			offsetx = sumPartition[j] * (*ldb);
		}


		if(!bupside && !btransA)
		{
                    offsetA = rowidx*ncolA
                              +colidx*tileheight;
		}else if (bupside && !btransA) {
                    offsetA = rowidx*ncolA
                              +colidx*tileheight;
		}else if( !bupside && btransA ){
                    offsetA = colidx*ncolA
                              +rowidx*tilewidth;
                }
                else if( bupside && btransA ){
                    offsetA = colidx*ncolA
                              +rowidx*tilewidth;
                }

		char transB = 'N';

/*   printf("i:%d,j:%d,offsetA:%d,offsetB:%d,offsetx:%d,tileheight:%d,tilewidth:%d\n",*/
/*                            i,j,offsetA,offsetB,offsetx,tileheight,tilewidth);*/

                if( i == j )
                {
                    /* B = A_{ii}^(-1)B */

                    gddgemm_host_multigpu_( &transB, transA, &nrowB,
                                           &tileheight, &tilewidth,
                                           alphah, alphal,
                                           hBh+offsetB, hBl+offsetB,
                                           ldb,
                                           hAh+offsetA, hAl+offsetA,
                                           &tileheight,
                                           &zeroh, &zerol,
                                           hBh+offsetB, hBl+offsetB,
                                           ldb );
                }
                else
                {
/*  printf("hAh:%e,hBhb:%e,bBhx:%e\n",*(hAh+offsetA+1),*/
		    if( !btransA )
		    {
			gddgemm_host_multigpu_( &transB,transA, &nrowB,
					       &tilewidth, &tileheight,
					       &moneh, &monel,
					       hBh+offsetx, hBl+offsetx,
					       ldb,
					       hAh+offsetA, hAl+offsetA,
					       &tileheight,
					       &oneh, &onel,
					       hBh+offsetB, hBl+offsetB,
					       ldb );
		    }
		    else
		    {
			gddgemm_host_multigpu_( &transB,transA, &nrowB,
					       &tilewidth, &tileheight,
					       &moneh, &monel,
					       hBh+offsetx, hBl+offsetx,
					       ldb,
					       hAh+offsetA, hAl+offsetA,
					       &tilewidth,
					       &oneh, &onel,
					       hBh+offsetB, hBl+offsetB,
					       ldb );
		    }
                }
            }
        }
    }

    if(bleftside)
    copyFromSplitMatrix( nrowB, ncolB, *ldb, ngpu, partition, hBh, hBl, Bh,
                         Bl );
    free( partition );
    free( sumPartition );
    cudaFreeHost( hAh );
    cudaFreeHost( hAl );
    if( bleftside )
    {
        cudaFreeHost( hBh );
        cudaFreeHost( hBl );
    }

    return 0;
}

#ifdef __cplusplus
}
#endif
