#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddzhemm_host_multigpu_( char* side, char* uplo,
                       int* m, int* n,
                       cuDoubleComplex* alphah, cuDoubleComplex* alphal,
                       cuDoubleComplex* Ah, cuDoubleComplex* Al, int* lda,
                       cuDoubleComplex* Bh, cuDoubleComplex* Bl, int* ldb,
                       cuDoubleComplex* betah, cuDoubleComplex* betal,
                       cuDoubleComplex* Ch, cuDoubleComplex* Cl, int* ldc )
{
    int      bleftside = 1;

    if( *side == 'R' || *side == 'r' ) bleftside = 0;
    /* NUMBER OF GPUS */
    int      ngpu;
    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int *    partitionL, *partitionR;
    int *    sumPartitionL, *sumPartitionR;
    partitionL = (int*) malloc( sizeof( int )*ngpu );
    sumPartitionL = (int*) malloc( sizeof( int )*ngpu );
    partitionR = (int*) malloc( sizeof( int )*ngpu );
    sumPartitionR = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *m, ngpu,  partitionL, sumPartitionL );
    CalcEachArraySize( *n, ngpu,  partitionR, sumPartitionR );

    int      sizeL = bleftside ? ( *m )*( *m ) : ( *m )*( *n );
    int      sizeR = bleftside ? ( *m )*( *n ) : ( *n )*( *n );
    int      sizeC = ( *m )*( *n );

    /* Working Matrix */


    cuDoubleComplex * hLh, *hLl, *hRh, *hRl, *hCh, *hCl;
    if( bleftside )
    {
        /* left side: C += A*B = hL*hR */
        if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hLh, sizeL*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hLl, sizeL*sizeof( cuDoubleComplex ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }

        copyToHermitianSplitMatrix( *m, *lda,uplo, ngpu, partitionL, Ah, Al, hLh, hLl );
	hRh = Bh;
	hRl = Bl;
    }
    else
    {
        /* right side side: C += B*A = hL*hR */

        if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hLh, sizeL*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hLl, sizeL*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hRh, sizeR*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hRl, sizeR*sizeof( cuDoubleComplex ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }
    copyToComplexSplitMatrix( *m, *n, *ldb, ngpu, partitionL, Bh, Bl, hLh, hLl );
    copyToHermitianMatrix(*n,*lda,uplo,Ah,Al,hRh,hRl);
    }


    if(
        cudaSuccess != cudaMallocHost( (void **)&hCh, sizeC*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&hCl, sizeC*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }
    copyToComplexSplitMatrix( *m, *n, *ldc, ngpu, partitionL, Ch, Cl, hCh, hCl );


    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        int     tid = omp_get_thread_num( );

        int     localldr;

        int     localn = partitionR[tid];
        if( bleftside )
        {
            localldr = *ldb;
        }
        else
        {
            localldr = *n;
        }

        int     lsizeR = localldr*localn;
        int     offsetR = localldr*sumPartitionR[tid];

        cuDoubleComplex *dRh, *dRl, *dalphah, *dalphal, *dbetah, *dbetal;

        /* B,alpha及びbetaを確保 */
        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeR, sizeof( cuDoubleComplex ), (void**) &dRh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( lsizeR, sizeof( cuDoubleComplex ), (void**) &dRl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetal )
          )
        {
            /*	 Memory allocation failed */
            cublasFree( dRh );
            cublasFree( dRl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            cublasFree( dbetah );
            cublasFree( dbetal );
            exit( -1 );
        }


        cublasSetVector( lsizeR, sizeof( cuDoubleComplex ), hRh+offsetR, 1, dRh, 1 );
        cublasSetVector( lsizeR, sizeof( cuDoubleComplex ), hRl+offsetR, 1, dRl, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphal, 1, dalphal, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), betah, 1, dbetah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), betal, 1, dbetal, 1 );


        int idx;
        for( idx = 0; idx < ngpu; idx++ )
        {
            int localm = partitionL[idx];
            int localldl=localm;
            int localk;

            if( bleftside )
            {
                localk = ( *m );
            }
            else
            {
                localk = ( *n );
            }

            int     offsetL = localk*sumPartitionL[idx];
            int     lsizeL    = localk*localm;

            int     localldc = localm;
            int     lsizeC = localm*localn;
            int     offsetC = sumPartitionR[tid]*localm+
                              sumPartitionL[idx]*( *n );

            cuDoubleComplex *dLh, *dLl, *dCh, *dCl;

            if(
                CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeL, sizeof( cuDoubleComplex ), (void**) &dLh )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeL, sizeof( cuDoubleComplex ), (void**) &dLl )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( cuDoubleComplex ), (void**) &dCh )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( cuDoubleComplex ), (void**) &dCl )
              )
            {
                /*	 Memory allocation failed */
                cublasFree( dLh );
                cublasFree( dLl );
                cublasFree( dCh );
                cublasFree( dCl );
                exit( -1 );
            }


            cublasSetVector( lsizeL, sizeof( cuDoubleComplex ), hLh+offsetL, 1, dLh,
                             1 );
            cublasSetVector( lsizeL, sizeof( cuDoubleComplex ), hLl+offsetL, 1, dLl,
                             1 );
            cublasSetVector( lsizeC, sizeof( cuDoubleComplex ), hCh+offsetC, 1, dCh,
                             1 );
            cublasSetVector( lsizeC, sizeof( cuDoubleComplex ), hCl+offsetC, 1, dCl,
                             1 );

            /* Execute gpu gemm*/
               char localtransL = 'N';
               char localtransR = 'N';

#ifdef _DEBUG_
            int devid;
            cudaGetDevice( &devid );
            printf( "TILE INDEX(%d,%d), Cuda Device = %d\n", idx, tid,
                    devid );
#endif

            gddzgemm( &localtransL, &localtransR,  &localm,  &localn,
                          &localk, dalphah,
                          dalphal,
                          dLh,  dLl,  &localldl, dRh,  dRl, &localldr,
                          dbetah, dbetal,
                          dCh,  dCl,
                          &localldc );
            cublasGetVector( lsizeC, sizeof( cuDoubleComplex ), dCh, 1, hCh+offsetC,
                             1 );
            cublasGetVector( lsizeC, sizeof( cuDoubleComplex ), dCl, 1, hCl+offsetC,
                             1 );

            cublasFree( dLh );
            cublasFree( dLl );
            cublasFree( dCh );
            cublasFree( dCl );
        }


        cublasFree( dRh );
        cublasFree( dRl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dbetah );
        cublasFree( dbetal );
    }
/*    for (int i = 0; i < *m*(*n); i++) {*/
/*            printf("%e\n",*(hCh+i));*/
/*    }*/

    copyFromComplexSplitMatrix( *m, *n, *ldc, ngpu, partitionL, hCh, hCl, Ch, Cl );

        cudaFreeHost( hLh );
        cudaFreeHost( hLl );
    if( !bleftside )
    {
        cudaFreeHost( hRh );
        cudaFreeHost( hRl );
    }
    cudaFreeHost( hCh );
    cudaFreeHost( hCl );

    free( partitionL );
    free( partitionR );
    free( sumPartitionL );
    free( sumPartitionR );

    return 0;
}

#ifdef __cplusplus
}
#endif
