#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddzher2_host_multigpu_( char* uplo, int *n, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
                      cuDoubleComplex * xh, cuDoubleComplex *xl,
                      int *incx,
                      cuDoubleComplex*yh, cuDoubleComplex*yl,
                      int *incy,
                      cuDoubleComplex *Ah, cuDoubleComplex *Al,
                      int *lda )
{
/* NUMBER OF GPUS */
    int      ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int *    partition;
    int *    pointerOfEachPartition;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );


    cuDoubleComplex *hAh,*hAl;
    int matsize = (*n)* (*n);
    if(
        cudaSuccess != cudaMallocHost(  (void **)&hAh, matsize*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&hAl, matsize*sizeof( cuDoubleComplex ) )
      )
    {
    cudaFreeHost( hAh );
    cudaFreeHost( hAl );
        printf( "Host Malloc Error\n" );
        return -1;
    }




    copyToHermitianMatrix(*n,*lda,uplo,Ah,Al,hAh,hAl);
    cuDoubleComplex conjalphah = cuConj(*alphah);
    cuDoubleComplex conjalphal = cuConj(*alphal);

    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        cuDoubleComplex *dxh, *dxl, *dyh, *dyl, *dAh, *dAl,*dalphah,*dalphal,*dconjalphah,*dconjalphal;

        int     tid = omp_get_thread_num( );
        int     sizex = ( *n )*( *incx );
        int     sizey = ( *n )*( *incy );
	int     offsetx = pointerOfEachPartition[tid]*( *incx );
        int     offsety = pointerOfEachPartition[tid]*( *incy );
	int     sizeA = (*n)*partition[tid];
	int     offsetA = pointerOfEachPartition[tid]*(*n);


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dconjalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dconjalphal )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dyh );
            cublasFree( dyl );
            cublasFree( dAh );
            cublasFree( dAl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( sizex, sizeof( cuDoubleComplex ), xh, 1, dxh,1 );
            cublasSetVector( sizex, sizeof( cuDoubleComplex ), xl, 1, dxl,1 );
            cublasSetVector( sizey, sizeof( cuDoubleComplex ), yh, 1, dyh,1 );
            cublasSetVector( sizey, sizeof( cuDoubleComplex ), yl, 1, dyl,1 );
            cublasSetVector( sizeA, sizeof( cuDoubleComplex ), hAh+offsetA, 1, dAh,1 );
            cublasSetVector( sizeA, sizeof( cuDoubleComplex ), hAl+offsetA, 1, dAl,1 );
            cublasSetVector( 1, sizeof( cuDoubleComplex ), alphah, 1, dalphah,1 );
            cublasSetVector( 1, sizeof( cuDoubleComplex ), alphal, 1, dalphal,1 );
            cublasSetVector( 1, sizeof( cuDoubleComplex ), &conjalphah, 1, dconjalphah,1 );
            cublasSetVector( 1, sizeof( cuDoubleComplex ), &conjalphal, 1, dconjalphal,1 );
        }


        /* Execute gpu vector add */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Cuda Device = %d\n", tid, devid );
#endif
        gddzgerc( n,&partition[tid], dalphah,dalphal, dxh, dxl, incx,
                     dyh+offsety, dyl+offsety, incy,
                     dAh,dAl,lda );

        gddzgerc( n,&partition[tid], dconjalphah,dconjalphal,
                     dyh, dyl, incy, dxh+offsetx, dxl+offsetx, incx,
                     dAh,dAl,lda );

        /* Get From Device Memory */
        cublasGetVector( sizeA, sizeof( cuDoubleComplex ), dAh, 1, hAh+offsetA, 1 );
        cublasGetVector( sizeA, sizeof( cuDoubleComplex ), dAl, 1, hAl+offsetA, 1 );

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dAh );
        cublasFree( dAl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dconjalphah );
        cublasFree( dconjalphal );
    }



    
    copyFromHermitianMatrix(*n,*lda,uplo,hAh,hAl,Ah,Al);

    free( partition );
    free( pointerOfEachPartition );
    cudaFreeHost( hAh );
    cudaFreeHost( hAl );
    return 0;
}

#ifdef __cplusplus
}
#endif
