#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddzher2k_host_multigpu_back_( char* uplo, char* trans,
                        int* n, int* k,
                        cuDoubleComplex* alphah, cuDoubleComplex* alphal,
                        cuDoubleComplex* Ah, cuDoubleComplex* Al, int* lda,
                        cuDoubleComplex* Bh, cuDoubleComplex* Bl, int* ldb,
                        double* betah, double* betal,
                        cuDoubleComplex* Ch, cuDoubleComplex* Cl, int* ldc )
{
    int      btrans = 0;

    if( *trans == 'T' || *trans == 't' ) btrans = 1;

    /* NUMBER OF GPUS */
    int      ngpu;
    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int *    partition;
    int *    sumPartition;
    partition = (int*) malloc( sizeof( int )*ngpu );
    sumPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, sumPartition );

    int      sizeM = ( *n )*( *k );
    int      sizeC = ( *n )*( *n );

    /* Working Matrix */


    cuDoubleComplex * hAh, *hAl, *hBh, *hBl, *hCh, *hCl;
    if( !btrans )
    {
        if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hAh, sizeM*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hAl, sizeM*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hBh, sizeM*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hBl, sizeM*sizeof( cuDoubleComplex ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }

        copyToComplexSplitMatrix( *n, *k, *lda, ngpu, partition, Ah, Al, hAh, hAl );
        copyComplexTranspose( *k, *n, *ldb, Bh, Bl, hBh, hBl );
    }
    else
    {
        hAh = Ah;
        hAl = Al;
        hBh = Bh;
        hBl = Bl;
    }


    if(
        cudaSuccess != cudaMallocHost( (void **)&hCh, sizeC*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&hCl, sizeC*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }
    copyToHermitianSplitMatrix( *n, *ldc, uplo, ngpu, partition, Ch, Cl,
                                hCh,hCl );


    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        int     tid = omp_get_thread_num( );

        int     localldb = btrans ? ( *ldb ) : ( *k );
        int     localn = partition[tid];

        cuDoubleComplex *dAh[ngpu], *dAl[ngpu], *dBh[ngpu], *dBl[ngpu], *dalphah, *dalphal,
        *dbetah, *dbetal,*conjdalphah,*conjdalphal,
        *doneh, *donel;

        /* B,alpha及びbetaを確保 */
        for( int i = 0; i < ngpu; i++ )
        {
            int sizeA = btrans ? ( *lda )*partition[i] : ( *k )*
                        partition[i];
            int sizeB = btrans ? ( *ldb )*partition[i] : ( *k )*
                        partition[i];

            if(
                CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAh[i] )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAl[i] )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeB, sizeof( cuDoubleComplex ), (void**) &dBh[i] )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( sizeB, sizeof( cuDoubleComplex ), (void**) &dBl[i] )
              )
            {
                exit( -1 );
            }
        }
        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &conjdalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &conjdalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &doneh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &donel )
          )
        {
            cublasFree( dalphah );
            cublasFree( dalphal );
            cublasFree( conjdalphah );
            cublasFree( conjdalphal );
            cublasFree( dbetah );
            cublasFree( dbetal );
            cublasFree( doneh );
            cublasFree( donel );
            exit( -1 );
        }

        cuDoubleComplex oneh = make_cuDoubleComplex(1.0,0.0);
        cuDoubleComplex onel = make_cuDoubleComplex(0.0,0.0);

        for( int i = 0; i < ngpu; i++ )
        {
            int offsetA = btrans ? ( *lda )*sumPartition[i] : ( *k )*
                          sumPartition[i];
            int offsetB = btrans ? ( *ldb )*sumPartition[i] : ( *k )*
                          sumPartition[i];
            int sizeA = btrans ? ( *lda )*partition[i] : ( *k )*
                        partition[i];
            int sizeB = btrans ? ( *ldb )*partition[i] : ( *k )*
                        partition[i];
            cublasSetVector( sizeA, sizeof( cuDoubleComplex ), hAh+offsetA, 1,
                             dAh[i],
                             1 );
            cublasSetVector( sizeA, sizeof( cuDoubleComplex ), hAl+offsetA, 1,
                             dAl[i],
                             1 );
            cublasSetVector( sizeB, sizeof( cuDoubleComplex ), hBh+offsetB, 1,
                             dBh[i],
                             1 );
            cublasSetVector( sizeB, sizeof( cuDoubleComplex ), hBl+offsetB, 1,
                             dBl[i],
                             1 );
        }

	cuDoubleComplex cbetah,cbetal,conjalphah,conjalphal;
	cbetah = make_cuDoubleComplex(*betah,0.0);
	cbetal = make_cuDoubleComplex(*betal,0.0);
	conjalphah = cuConj(*alphah);
	conjalphal = cuConj(*alphal);
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphal, 1, dalphal, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), &conjalphah, 1, conjdalphah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), &conjalphal, 1, conjdalphal, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), &cbetah, 1, dbetah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), &cbetal, 1, dbetal, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), &oneh, 1, doneh, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), &onel, 1, donel, 1 );

        int idx;
        for( idx = 0; idx < ngpu; idx++ )
        {
            int     localm = partition[idx];
            int     localk = *k;


            int     localldc = localm;
            int     lsizeC = localm*localn;
            int     offsetC = sumPartition[tid]*localm+
                              sumPartition[idx]*( *n );

            cuDoubleComplex *dCh, *dCl;

            if(
                CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( cuDoubleComplex ), (void**) &dCh )
              || CUBLAS_STATUS_SUCCESS !=
                cublasAlloc( lsizeC, sizeof( cuDoubleComplex ), (void**) &dCl )
              )
            {
                /*	 Memory allocation failed */
                cublasFree( dCh );
                cublasFree( dCl );
                exit( -1 );
            }


            cublasSetVector( lsizeC, sizeof( cuDoubleComplex ), hCh+offsetC, 1, dCh,
                             1 );
            cublasSetVector( lsizeC, sizeof( cuDoubleComplex ), hCl+offsetC, 1, dCl,
                             1 );

#ifdef _DEBUG_
            int  devid;
            cudaGetDevice( &devid );
            printf( "TILE INDEX(%d,%d), Cuda Device = %d\n", idx, tid,
                    devid );
#endif


            char localtransA = btrans ? 'C' : 'N';
            char localtransB = btrans ? 'N' : 'N';

            int  locallda = btrans ? ( *lda ) : partition[idx];

            /* Execute gpu gemm*/

            gddzgemm( &localtransA, &localtransB,  &localm,  &localn,
                          &localk, dalphah,
                          dalphal,
                          dAh[idx],  dAl[idx],  &locallda, dBh[tid],
                          dBl[tid], &localldb,
                          dbetah, dbetal,
                          dCh,  dCl,
                          &localldc );

            localtransA = btrans ? 'N' : 'C';
            localtransB = btrans ? 'C' : 'C';
            locallda = btrans ? ( *lda ) : partition[tid];

            gddzgemm( &localtransB, &localtransA,  &localm,  &localn,
                          &localk, conjdalphah,
                          conjdalphal,
                          dBh[idx],  dBl[idx],  &localldb, dAh[tid],
                          dAl[tid], &locallda,
                          doneh, donel,
                          dCh,  dCl,
                          &localldc );


            cublasGetVector( lsizeC, sizeof( cuDoubleComplex ), dCh, 1, hCh+offsetC,
                             1 );
            cublasGetVector( lsizeC, sizeof( cuDoubleComplex ), dCl, 1, hCl+offsetC,
                             1 );

            cublasFree( dCh );
            cublasFree( dCl );
        }

        for( int i = 0; i < ngpu; i++ )
        {
            cublasFree( dAh[i] );
            cublasFree( dAl[i] );
        }

        cublasFree( dBh );
        cublasFree( dBl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dbetah );
        cublasFree( dbetal );
        cublasFree( doneh );
        cublasFree( donel );
    }


    copyFromHermitianSplitMatrix( *n, *ldc, uplo, ngpu, partition, hCh, hCl,Ch, Cl );


    if( !btrans )
    {
        cudaFreeHost( hAh );
        cudaFreeHost( hAl );
        cudaFreeHost( hBh );
        cudaFreeHost( hBl );
    }
    cudaFreeHost( hCh );
    cudaFreeHost( hCl );

    free( partition );
    free( sumPartition );

    return 0;
}

#ifdef __cplusplus
}
#endif
